/* Parallel Hierarchical Grid -- an adaptive finite element library.
 *
 * Copyright (C) 2005-2007 State Key Laboratory of Scientific and
 * Engineering Computing, Chinese Academy of Sciences. */

/* This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA */

/* $Id: dg.c,v 1.21 2007/05/22 07:50:47 zlb Exp $ */

/* This file implements the discontinuous Galerkin elements */

#include "phg.h"

static const FLOAT *
DG_bas(DOF *dof, SIMPLEX *e, int no0, int no1, const FLOAT *lambda)
{
    DOF hb;

    if (dof->type->order < 5)
	return DOF_Pn[dof->type->order]->BasFuncs(dof, e, no0, no1, lambda);
    hb.g = dof->g;
    hb.type = DOF_HBn[dof->type->order];
    hb.dim = dof->dim;
    return hb.type->BasFuncs(&hb, e, no0, no1, lambda);
}

static const FLOAT *
DG_grad(DOF *dof, SIMPLEX *e, int no0, int no1, const FLOAT *lambda)
{
    DOF hb;

    if (dof->type->order < 5)
	return DOF_Pn[dof->type->order]->BasGrads(dof, e, no0, no1, lambda);

    hb.g = dof->g;
    hb.type = DOF_HBn[dof->type->order];
    hb.dim = dof->dim;
    return hb.type->BasGrads(&hb, e, no0, no1, lambda);
}

#if 0
static void
DG_interp(DOF *dof, SIMPLEX *e, FLOAT **old_data, FLOAT **new_data)
/* interpolation */
{
    /* TODO: permute edge and face data, then call lagrange->Pn_interp */
}
#endif

static FLOAT DG0_points[] = {0.25L,0.25L,0.25L,0.25L};
DOF_TYPE DOF_DG0_ = {DofCache, "DG0", DG0_points, NULL, DOF_DG0,
		     phgDofInterpC2FGeneric, phgDofInterpF2CGeneric,
		     phgDofInitFuncPoint,
		     DG_bas, DG_grad,
		     TRUE, FALSE, -1, 1, 0, -1, 1,
		     0, 0, 0, 1};

static FLOAT DG1_points[] = {
    1.L,0.L,0.L,0.L,  0.L,1.L,0.L,0.L,  0.L,0.L,1.L,0.L,  0.L,0.L,0.L,1.L
};
DOF_TYPE DOF_DG1_ = {DofCache, "DG1", DG1_points, NULL, DOF_DG0,
		     phgDofInterpC2FGeneric, phgDofInterpF2CGeneric,
		     phgDofInitFuncPoint,
		     DG_bas, DG_grad,
		     TRUE, FALSE, -1, NVert, 1, -1, 1,
		     0, 0, 0, NVert};

static FLOAT DG2_points[] = {
    1.L,0.L,0.L,0.L,  0.L,1.L,0.L,0.L,  0.L,0.L,1.L,0.L,  0.L,0.L,0.L,1.L,

    .5L,.5L,0.L,0.L,  .5L,0.L,.5L,0.L,  .5L,0.L,0.L,.5L,
    0.L,.5L,.5L,0.L,  0.L,.5L,0.L,.5L,  0.L,0.L,.5L,.5L
};
DOF_TYPE DOF_DG2_ = {DofCache, "DG2", DG2_points, NULL, DOF_DG1,
		     phgDofInterpC2FGeneric, phgDofInterpF2CGeneric,
		     phgDofInitFuncPoint,
		     DG_bas, DG_grad,
		     TRUE, FALSE, -1, NVert + NEdge, 2, -1, 1,
		     0, 0, 0, NVert + NEdge};

static FLOAT DG3_points[] = {
    1.L,0.L,0.L,0.L,  0.L,1.L,0.L,0.L,  0.L,0.L,1.L,0.L,  0.L,0.L,0.L,1.L,

    2.L/3.L, 1.L/3.L, 0.L, 0.L,   1.L/3.L, 2.L/3.L, 0.L, 0.L,
    2.L/3.L, 0.L, 1.L/3.L, 0.L,   1.L/3.L, 0.L, 2.L/3.L, 0.L,
    2.L/3.L, 0.L, 0.L, 1.L/3.L,   1.L/3.L, 0.L, 0.L, 2.L/3.L,
    0.L, 2.L/3.L, 1.L/3.L, 0.L,   0.L, 1.L/3.L, 2.L/3.L, 0.L,
    0.L, 2.L/3.L, 0.L, 1.L/3.L,   0.L, 1.L/3.L, 0.L, 2.L/3.L,
    0.L, 0.L, 2.L/3.L, 1.L/3.L,   0.L, 0.L, 1.L/3.L, 2.L/3.L,

    0.L, 1.L/3.L, 1.L/3.L, 1.L/3.L,
    1.L/3.L, 0.L, 1.L/3.L, 1.L/3.L,
    1.L/3.L, 1.L/3.L, 0.L, 1.L/3.L,
    1.L/3.L, 1.L/3.L, 1.L/3.L, 0.L
};
DOF_TYPE DOF_DG3_ = {DofCache, "DG3", DG3_points, NULL, DOF_DG2,
		     phgDofInterpC2FGeneric, phgDofInterpF2CGeneric,
		     phgDofInitFuncPoint,
		     DG_bas, DG_grad,
		     TRUE, FALSE, -1, NVert + 2*NEdge + NFace, 3, -1, 1,
		     0, 0, 0, NVert + 2*NEdge + NFace};

static FLOAT DG4_points[] = {
    1.L,0.L,0.L,0.L,  0.L,1.L,0.L,0.L,  0.L,0.L,1.L,0.L,  0.L,0.L,0.L,1.L,

    .75L, .25L, 0.L, 0.L,   .5L, .5L, 0.L, 0.L,   .25L, .75L, 0.L, 0.L,
    .75L, 0.L, .25L, 0.L,   .5L, 0.L, .5L, 0.L,   .25L, 0.L, .75L, 0.L,
    .75L, 0.L, 0.L, .25L,   .5L, 0.L, 0.L, .5L,   .25L, 0.L, 0.L, .75L,
    0.L, .75L, .25L, 0.L,   0.L, .5L, .5L, 0.L,   0.L, .25L, .75L, 0.L,
    0.L, .75L, 0.L, .25L,   0.L, .5L, 0.L, .5L,   0.L, .25L, 0.L, .75L,
    0.L, 0.L, .75L, .25L,   0.L, 0.L, .5L, .5L,   0.L, 0.L, .25L, .75L,

    0.L, .5L, .25L, .25L,    0.L, .25L, .5L, .25L,   0.L, .25L, .25L, .5L,
    .5L, 0.L, .25L, .25L,    .25L, 0.L, .5L, .25L,   .25L, 0.L, .25L, .5L,
    .5L, .25L, 0.L, .25L,    .25L, .5L, 0.L, .25L,   .25L, .25L, 0.L, .5L,
    .5L, .25L, .25L, 0.L,    .25L, .5L, .25L, 0.L,   .25L, .25L, .5L, 0.L,

    .25L, .25L, .25L, .25L
};
DOF_TYPE DOF_DG4_ = {DofCache, "DG4", DG4_points, NULL, DOF_DG3,
		     phgDofInterpC2FGeneric, phgDofInterpF2CGeneric,
		     phgDofInitFuncPoint,
		     DG_bas, DG_grad,
		     TRUE, FALSE, -1, NVert + 3*(NEdge + NFace) + 1, 4, -1, 1,
		     0, 0, 0, NVert + 3*(NEdge + NFace) + 1};

DOF_TYPE *DOF_DGn[] = {
    DOF_DG0,
    DOF_DG1, DOF_DG2, DOF_DG3, DOF_DG4,
    DOF_DG5, DOF_DG6, DOF_DG7, DOF_DG8, DOF_DG9, DOF_DG10,
    DOF_DG11, DOF_DG12, DOF_DG13, DOF_DG14, DOF_DG15,
    NULL
};

static void
DG_init(DOF *dof, SIMPLEX *e, GTYPE type, int index,
	DOF_USER_FUNC userfunc,
	DOF_USER_FUNC_LAMBDA userfunc_lambda,
	const FLOAT *funcvalues, FLOAT *dofvalues, FLOAT **pdofvalues)
{
    DOF_HB1->InitFunc(dof, e, type, index, userfunc, userfunc_lambda,
			funcvalues, dofvalues, pdofvalues);
}

/* DGp = Hierarchical basis of order p, p > 4 */
#define NBas(p)	NVert + (p-1) * NEdge + (p-1)*(p-2)/2 * NFace + \
		(p-1)*(p-2)*(p-3)/6

#define DofType(order, type, name, grad)				\
    DOF_TYPE type = {DofCache, name, NULL, NULL, grad,			\
		     phgDofInterpC2FGeneric, phgDofInterpF2CGeneric,	\
		     DG_init, DG_bas, DG_grad,				\
		     FALSE, FALSE, -1, NBas(order), order, -1, 1,	\
		     0, 0, 0, NBas(order)};
DofType(5,  DOF_DG5_,  "DG5",  DOF_DG4);
DofType(6,  DOF_DG6_,  "DG6",  DOF_DG5);
DofType(7,  DOF_DG7_,  "DG7",  DOF_DG6);
DofType(8,  DOF_DG8_,  "DG8",  DOF_DG7);
DofType(9,  DOF_DG9_,  "DG9",  DOF_DG8);
DofType(10, DOF_DG10_, "DG10", DOF_DG9);
DofType(11, DOF_DG11_, "DG11", DOF_DG10);
DofType(12, DOF_DG12_, "DG12", DOF_DG11);
DofType(13, DOF_DG13_, "DG13", DOF_DG12);
DofType(14, DOF_DG14_, "DG14", DOF_DG13);
DofType(15, DOF_DG15_, "DG15", DOF_DG14);
