/* Parallel Hierarchical Grid -- an adaptive finite element library.
 *
 * Copyright (C) 2005-2007 State Key Laboratory of Scientific and
 * Engineering Computing, Chinese Academy of Sciences. */

/* This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA */

/* $Id: dof.c,v 1.136 2007/05/11 01:54:38 zlb Exp $ */

#include "phg.h"
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <stdarg.h>

DOF_TYPE *DOF_DEFAULT = DOF_P2;

int
phgDofBasisOrder(DOF *u, int ibas)
{
    DOF_TYPE *type = u->type;
    BYTE *orders;
    int n, m;

    if ((orders = type->orders) == NULL || ibas < 0 || ibas >= type->nbas)
	return type->order;
    if (ibas < (n = NVert * type->np_vert))
	return orders[ibas % type->np_vert];
    orders += type->np_vert;
    if ((ibas -= n) < (m = NEdge * type->np_edge))
	return orders[ibas % type->np_edge];
    orders += type->np_edge;
    if ((ibas -= m) < (n = NFace * type->np_face))
	return orders[ibas % type->np_face];
    return orders[type->np_face + ibas - n];
}

/* Special DOFs */

DOF_TYPE DOF_CONSTANT_ = {DofCache, "Constant", NULL, NULL, &DOF_CONSTANT_,
    NULL, NULL, NULL, NULL, NULL,
    TRUE, FALSE, -1,		/* invariant, free_after_use, id */
    0, 0, 0, 1,			/* nbas, order, cont, dim */
    0, 0, 0, 0
};				/* np_vert, np_edge, np_face, np_elem */

DOF_TYPE DOF_ANALYTIC_ = {DofCache, "Analytic", NULL, NULL, NULL,
    NULL, NULL, NULL, NULL, NULL,
    FALSE, FALSE, -1,		/* invariant, free_after_use, id */
    0, -1, 0, 1,		/* nbas, order, cont, dim */
    0, 0, 0, 0
};				/* np_vert, np_edge, np_face, np_elem */

/* Dummy USER_FUNCs */

void
phgDofUserFuncNoData_(FLOAT x, FLOAT y, FLOAT z, FLOAT *values)
{
    Unused(x);
    Unused(y);
    Unused(z);
    Unused(values);
    return;
}

void
phgDofUserFuncNoAction_(FLOAT x, FLOAT y, FLOAT z, FLOAT *values)
{
    Unused(x);
    Unused(y);
    Unused(z);
    Unused(values);
    return;
}

void
phgDofUserFuncInterpolation_(FLOAT x, FLOAT y, FLOAT z, FLOAT *values)
{
    Unused(x);
    Unused(y);
    Unused(z);
    Unused(values);
    return;
}

/* list of referenced DOF_TYPES */
static struct {
    DOF_TYPE *type;
    size_t refcount;
} *type_list = NULL;
static size_t type_list_count = 0, type_list_allocated = 0;

DOF *
phgDofNew_(GRID *g, DOF_TYPE *type, SHORT dim, const char *name,
	   DOF_USER_FUNC userfunc, const char *srcfile, int srcline)
{
    DOF *dof;
    size_t len;

    FunctionEntry;

    if (phgMasterSlave)		/* Master/slave mode is not yet implemented */
	phgError(1, "phgDofNew does not allow master/slave mode\n");

    if (!phgGridIsValid(g))
	phgError(1, "%s:%d, invalid grid.\n", srcfile == NULL ? "" : srcfile,
			srcline);

    dof = phgCalloc(1, sizeof(*dof));
    dof->magic = MAGIC_DOF;
    if (name == NULL)
	name = "noname";
    phgInfo(2, "creating DOF \"%s\".\n", name);
    dof->name = phgAlloc(strlen(name) + 1);
    strcpy(dof->name, name);
    dof->type = type;
    dof->userfunc = userfunc;
    dof->invariant = FALSE;
    dof->DB_mask = DIRICHLET;
    if (type == DOF_CONSTANT && userfunc != DofNoAction
	&& userfunc != DofNoData) {
	phgPrintf("changing userfunc of constant DOF \"%s\" to DofNoAction\n",
		  dof->name);
	dof->userfunc = DofNoAction;
    }
    if (userfunc == DofInterpolation && type->InterpC2F == NULL) {
	phgPrintf("DOF \"%s\": no interpolation function for type \"%s\", "
		  "changing userfunc to DofNoAction\n", dof->name,
		  type->name);
	dof->userfunc = DofNoAction;
    }
    dof->g = g;
    if ((dof->dim = dim) <= 0)
	phgError(1, "invalid value for dim (%d)", dim);
    dof->srcfile = srcfile;
    dof->srcline = srcline;

    if (!SpecialDofType(type)) {
	if (type->np_vert > 0 && !g->vert_flag)
	    phgError(1, "%s elements has vertex DOFs "
		     "but VERT_FLAG is FALSE.\n", dof->type->name);
	if (type->np_edge > 0 && !g->edge_flag)
	    phgError(1, "%s elements require edge indices "
		     "but EDGE_FLAG is FALSE.\n", dof->type->name);
	if (type->np_face > 0 && !g->face_flag)
	    phgError(1, "%s elements require face indices "
		     "but FACE_FLAG is FALSE.\n", dof->type->name);
	if (type->np_elem > 0 && !g->elem_flag)
	    phgError(1, "%s elements require element indices "
		     "but ELEM_FLAG is FALSE.\n", dof->type->name);
    }

    /* attach the new DOF to g */
    len = 0;
    if (g->dof != NULL) {
	DOF **p;
	for (p = g->dof; *p != NULL; p++, len++);
    }
    g->dof = phgRealloc(g->dof, (len + 2) * sizeof(*(g->dof)));
    g->dof[len] = dof;
    g->dof[len + 1] = NULL;

    /* init dof's cache */
    dof->cache_func = NULL;

    dof->count_vert = dim * (SpecialDofType(type) ? 0 : type->np_vert);
    dof->count_edge = dim * (SpecialDofType(type) ? 0 : type->np_edge);
    dof->count_face = dim * (SpecialDofType(type) ? 0 : type->np_face);
    dof->count_elem = dim * (SpecialDofType(type) ? 0 : type->np_elem);

    /* init DOF values */
    if (dof->userfunc == DofNoData) {
	dof->data = NULL;
	dof->userfunc = DofNoAction;
    }
    else {
	phgDofSetDataByFunction(dof, userfunc);
    }

    /* update the list of referenced DOF_TYPE's */
    if (type->id >= 0) {
	assert(type->id < type_list_count
	       && type_list[type->id].type == type);
	type_list[type->id].refcount++;
    }
    else {
	/* add a new entry */
	if (type_list_count >= type_list_allocated) {
	    len = (type_list_allocated += 8) * sizeof(*type_list);
	    type_list = phgRealloc(type_list, len);
	}
	type_list[type_list_count].type = type;
	type_list[type_list_count].refcount = 1;
	type->id = type_list_count++;
    }

    Return dof;
    return dof;		/* to make MIPS cc happy */
}

void
phgDofFree(DOF **dof_ptr)
{
    DOF *dof = (dof_ptr == NULL ? NULL : *dof_ptr);

    MagicCheck(DOF, dof)

    if (dof == NULL)
	return;

    if (dof->g != NULL && !phgDofIsValid(dof->g, dof))
	phgError(1, "phgDofFree: invalid DOF object.\n");

    phgInfo(2, "freeing DOF \"%s\".\n", 
		dof->name == NULL ? "nonname" : dof->name);

    if (dof->type->id >= 0) {
	/* update the list of referenced DOF_TYPEs */
	assert(dof->type->id < type_list_count &&
	       type_list[dof->type->id].type == dof->type);
	if (--type_list[dof->type->id].refcount == 0) {
	    /* remove the type from type_list[] */
	    phgDofClearCache(NULL, NULL, dof->type, TRUE);
	    if (dof->type->id < --type_list_count) {
		int i;
		memmove(type_list + dof->type->id,
			type_list + dof->type->id + 1,
			(type_list_count -
			 dof->type->id) * sizeof(*type_list));
		for (i = dof->type->id; i < type_list_count; i++)
		    type_list[i].type->id = i;
	    }
	    if (dof->type->free_after_use) {
		phgFree((void *)dof->type->name);
		phgFree(dof->type->points);
		phgFree(dof->type);
	    }
	    else
		dof->type->id = -1;
	    if (type_list_count == 0) {
		phgFree(type_list);
		type_list = NULL;
		type_list_allocated = 0;
	    }
	}
    }

    phgFree(dof->name);
    phgFree(dof->data);
    phgDofClearCache(NULL, dof, NULL, TRUE);

    /* detach the dof from the mesh */
    if (phgGridIsValid(dof->g)) {
	DOF **p, **p0 = NULL;
	size_t len = 0;
	for (p = dof->g->dof; p != NULL && *p != NULL; p++, len++) {
	    if (*p == dof) {
		len = 0;
		p0 = p;
	    }
	}
	if (p == NULL || p0 == NULL)
	    phgError(1, "%s:%d: unexpected error (grid freed before DOF?)\n",
		     __FILE__, __LINE__);
	memmove(p0, p0 + 1, sizeof(*p0) * len);
    }

    phgFree(dof);

    *dof_ptr = NULL;

    return;
}

BTYPE
phgDofSetDirichletBoundaryMask(DOF *u, BTYPE mask)
{
    BTYPE old_mask = u->DB_mask;
    u->DB_mask = mask;
    return old_mask;
}

BOOLEAN
phgDofIsValid(GRID *g, DOF *dof)
{
    DOF **p;

    if (g == NULL || (p = g->dof) == NULL)
	return FALSE;

    while (*p != dof && *p != NULL)
	p++;

    return *p == dof;
}

DOF *
phgDofIndex2Pointer(GRID *g, int id)
{
    DOF **p;

    if ((p = g->dof) == NULL)
	return NULL;

    while (id-- > 0 && *p != NULL) {
	p++;
    }

    return *p;
}

int
phgDofPointer2Index(DOF *dof)
{
    DOF **p;

    if (dof == NULL)
	return -1;

    p = dof->g->dof;
    while (*p != NULL && *p != dof) {
	p++;
    }

    return (*p == dof) ? p - dof->g->dof : -1;
}

static void
clear_cache0(DOF_TYPE *type, BOOLEAN final)
{
    if (final || !type->invariant) {
	phgQuadClearDofCache(&type->cache_basfunc, NULL, final);
	phgQuadClearDofCache(&type->cache_basgrad, NULL, final);
    }
    phgQuadClearDofCache(&type->cache_gradient, NULL, final);
    phgQuadClearDofCache(&type->cache_curl, NULL, final);
}

void
phgDofClearCache(GRID *g, DOF *dof, DOF_TYPE *type, BOOLEAN final)
/* when 'type' is NULL and dof is not NULL, 
 * when dof is not NULL , the parameter 'g' take no effective and the cache in 'type' will be cleared. In this case, none cache in referenced DOF_TYPE will be cleared.
 * when dof is  NULL, one or all of the referenced DOF's caches or DOF_TYPE's caches will be cleared. */
{
    size_t i;
    DOF **dof1;

    if (dof != NULL) {
	phgQuadClearDofCache(&(dof->cache_func), NULL, final);
	if (type != NULL) {
	    clear_cache0(type, final);
	    return;
	}
    }

    if (g != NULL && g->dof != NULL) {
	for (dof1 = g->dof; *dof1 != NULL; dof1++)
	    phgQuadClearDofCache(&((*dof1)->cache_func), NULL, final);
    }

    if (type != NULL) {
	clear_cache0(type, final);
	return;
    }
    else {
	/* clear all quad caches */
	for (i = 0; i < type_list_count; i++)
	    clear_cache0(type_list[i].type, final);
    }
}

void
phgDofMapEdgeData(const DOF_TYPE *type, const SIMPLEX *e, int no, int *M)
/* returns in M[] ordering of data according to edge direction */
{
    int i;

    if (type->np_edge == 0)
	return;

    if (type->np_edge <= 1 || type->BasFuncs == NULL || type->points == NULL
	|| MapVertex(e, GetEdgeVertex(no, 0))
	< MapVertex(e, GetEdgeVertex(no, 1))) {
	for (i = 0; i < type->np_edge; i++)
	    M[i] = i;
    }
    else {
	for (i = 0; i < type->np_edge; i++)
	    M[i] = type->np_edge - 1 - i;
    }

    return;
}

void
phgDofMapFaceData(const DOF_TYPE *type, const SIMPLEX *e, int no, int *M)
/* compute ordering of data according to global indices of face vertices */
{
    int v0, v1, v2;

    if (type->np_face <= 1 || type->BasFuncs == NULL || type->points == NULL) {
	for (v0 = 0; v0 < type->np_face; v0++)
	    M[v0] = v0;
	return;
    }
    else if (type->np_face != 3) {
	/* TODO: Decompose 'np_face' := 6 * n + 3 * m + l (l = 0 or 1),
	   if 'points' := n S111 + m S211 + l S3.
	   Sxxx denotes the set of points obtained by permuting xxx. */
	phgError(1, "%s:%d: unimplemented case (np_face=%d)!\n",
		 __FILE__, __LINE__, type->np_face);
    }

    M[0] = 0;
    M[1] = 0;
    M[2] = 0;
    v0 = MapVertex(e, GetFaceVertex(no, 0));
    v1 = MapVertex(e, GetFaceVertex(no, 1));
    v2 = MapVertex(e, GetFaceVertex(no, 2));
    (v0 > v1) ? M[0]++ : M[1]++;
    (v0 > v2) ? M[0]++ : M[2]++;
    (v1 > v2) ? M[1]++ : M[2]++;

    return;
}

static struct {
    SIMPLEX *e;			/* the element having the DOF */
    BYTE index;			/* local index of the DOF */
    BYTE flag;			/* flag */
} *tmp, *tmp_vert, *tmp_edge, *tmp_face, *tmp_elem, *tmp_end;
static BOOLEAN flagged_only = FALSE;

static BOOLEAN
  setup_tmp_callback
CB_ARGS(e)
{
    int i;

    if (tmp_vert < tmp_edge) {
	for (i = 0; i < NVert; i++) {
	    tmp = tmp_vert + e->verts[i];
	    if (tmp->flag && (!flagged_only || tmp->e->flag < 1))
		continue;
	    tmp->flag = TRUE;
	    tmp->e = e;
	    tmp->index = i;
	}
    }

    if (tmp_edge < tmp_face) {
	for (i = 0; i < NEdge; i++) {
	    tmp = tmp_edge + e->edges[i];
	    if (tmp->flag && (!flagged_only || tmp->e->flag < 1))
		continue;
	    tmp->flag = TRUE;
	    tmp->e = e;
	    tmp->index = i;
	}
    }

    if (tmp_face < tmp_elem) {
	for (i = 0; i < NFace; i++) {
	    tmp = tmp_face + e->faces[i];
	    if (tmp->flag && (!flagged_only || tmp->e->flag < 1))
		continue;
	    tmp->flag = TRUE;
	    tmp->e = e;
	    tmp->index = i;
	}
    }

    if (tmp_elem < tmp_end) {
	tmp = tmp_elem + e->index;
	if (tmp->flag)
	    return TRUE;
	tmp->flag = TRUE;
	tmp->e = e;
	tmp->index = 0;
    }

    return TRUE;
}

static void
setup_tmp(DOF *dof)
{
    size_t size0 = (dof->type->np_vert > 0) ? dof->g->nvert : 0,
	size1 = (dof->type->np_edge > 0) ? dof->g->nedge : 0,
	size2 = (dof->type->np_face > 0) ? dof->g->nface : 0,
	size3 = (dof->type->np_elem > 0) ? dof->g->nelem : 0;

    if (tmp_vert != NULL)
	phgFree(tmp_vert);
    tmp_vert = phgCalloc(size0 + size1 + size2 + size3, sizeof(*tmp_vert));
    tmp_edge = tmp_vert + size0;
    tmp_face = tmp_edge + size1;
    tmp_elem = tmp_face + size2;
    tmp_end = tmp_elem + size3;

    if (tmp_end > tmp_vert)
	phgTraverseElements(dof->g, setup_tmp_callback);
}

void
phgDofDump(DOF *dof)
{
    char s[1024], *fmt;
    int i, j, k, index = 0;
    GRID *g;
    FLOAT *data, *points = NULL, x, y, z;
    SIMPLEX *e = NULL;
    COORD *p0 = NULL, *p1 = NULL, *p2 = NULL, *p3 = NULL;
    static int *M = NULL, M_size = 0;

    if (dof == NULL)
	return;

    FreeAtExit(M);

    if (dof->type == DOF_ANALYTIC) {
	phgInfo(-1, "Analytic DOF \"%s\", dim = %d, function addr: %p\n",
		dof->name, dof->dim, dof->userfunc);
	return;
    }
    else if (dof->type == DOF_CONSTANT) {
	if (dof->data == NULL) {
	    phgInfo(-1, "Constant DOF \"%s\", dim = %d, no data\n",
		    dof->name, dof->dim);
	    return;
	}
	phgInfo(-1, "Constant DOF \"%s\", dim = %d, values:\n",
		dof->name, dof->dim);
	for (i = 0, j = 0; i < dof->dim / 4; i++, j += 4) {
	    phgInfo(-1, "    %10.4le %10.4le %10.4le %10.4le\n",
		    (double)dof->data[j], (double)dof->data[j + 1],
		    (double)dof->data[j + 2], (double)dof->data[j + 3]);
	}
	switch (dof->dim % 4) {
	    case 1:
		phgInfo(-1, "    %10.4le\n", (double)dof->data[j]);
		break;
	    case 2:
		phgInfo(-1, "    %10.4le %10.4le\n", (double)dof->data[j],
			(double)dof->data[j + 1]);
		break;
	    case 3:
		phgInfo(-1, "    %10.4le %10.4le %10.4le\n",
			(double)dof->data[j], (double)dof->data[j + 1],
			(double)dof->data[j + 2]);
		break;
	}
	return;
    }

    phgInfo(-1, "DOF \"%s\": dim=%dx%d, "
	    "np_vert=%d, np_edge=%d, np_face=%d, np_elem=%d\n",
	    dof->name, dof->dim, dof->type->dim,
	    dof->type->np_vert, dof->type->np_edge,
	    dof->type->np_face, dof->type->np_elem);
    if ((g = dof->g) == NULL || (data = dof->data) == NULL) {
	phgInfo(-1, "no data.\n");
	return;
    }

    if (dof->type->points != NULL) {
	int np_vert = dof->type->np_vert;	/* backup dof->type->np_vert */
	dof->type->np_vert = 0;	/* don't need tmp_vert */
	setup_tmp(dof);
	dof->type->np_vert = np_vert;	/* restore dof->type->np_vert */
    }

    if ((k = dof->type->np_vert) > 0) {
	fmt = (k <= 1) ? "%5d  " : ((k < 10) ? "%5d[%d]  " : "%5d[%02d]  ");
	phgInfo(-1, "Vertices:\n");
	for (i = 0; i < g->nvert; i++) {
	    for (j = 0; j < dof->type->np_vert; j++) {
		sprintf(s, fmt, GlobalVertex(g, i), j);
		if (dof->type->points != NULL)
		    sprintf(s + strlen(s), "%s(%6.3lf,%6.3lf,%6.3lf) = ",
			    dof->name, (double)g->verts[i][0],
			    (double)g->verts[i][1], (double)g->verts[i][2]);
		for (k = 0; k < dof->dim; k++)
		    sprintf(s + strlen(s), "%10.4le ", (double)*(data++));
		phgInfo(-1, "%s\n", s);
	    }
	}
    }

    if ((k = dof->type->np_edge) > 0) {
	if (M_size < k)
	    M = phgAlloc((M_size = k) * sizeof(*M));
	fmt = (k <= 1) ? "%5d  " : ((k < 10) ? "%5d[%d]  " : "%5d[%02d]  ");
	phgInfo(-1, "Edges:\n");
	for (i = 0; i < g->nedge; i++) {
	    if (dof->type->points != NULL) {
		if ((e = tmp_edge[i].e) == NULL) {
		    data += dof->count_edge;
		    continue;
		}
		index = tmp_edge[i].index;
		p0 = g->verts + e->verts[GetEdgeVertex(index, 0)];
		p1 = g->verts + e->verts[GetEdgeVertex(index, 1)];
		points = dof->type->points + dof->type->np_vert;
	    }
	    phgDofMapEdgeData(dof->type, e, index, M);
	    for (j = 0; j < dof->type->np_edge; j++) {
		sprintf(s, fmt, GlobalEdge(g, i), j);
		if (dof->type->points != NULL) {
		    x = (*p0)[0] * points[0] + (*p1)[0] * points[1];
		    y = (*p0)[1] * points[0] + (*p1)[1] * points[1];
		    z = (*p0)[2] * points[0] + (*p1)[2] * points[1];
		    sprintf(s + strlen(s), "%s(%6.3lf,%6.3lf,%6.3lf) = ",
			    dof->name, (double)x, (double)y, (double)z);
		    points += 2;
		}
		for (k = 0; k < dof->dim; k++)
		    sprintf(s + strlen(s), "%10.4le ",
			    (double)*(data + dof->dim * M[j] + k));
		phgInfo(-1, "%s\n", s);
	    }
	    data += dof->count_edge;
	}
    }

    if ((k = dof->type->np_face) > 0) {
	if (M_size < k)
	    M = phgAlloc((M_size = k) * sizeof(*M));
	fmt = (k <= 1) ? "%5d  " : ((k < 10) ? "%5d[%d]  " : "%5d[%02d]  ");
	phgInfo(-1, "Faces:\n");
	for (i = 0; i < g->nface; i++) {
	    if (dof->type->points != NULL) {
		if ((e = tmp_face[i].e) == NULL) {
		    data += dof->count_face;
		    continue;
		}
		index = tmp_face[i].index;
		p0 = g->verts + e->verts[GetFaceVertex(index, 0)];
		p1 = g->verts + e->verts[GetFaceVertex(index, 1)];
		p2 = g->verts + e->verts[GetFaceVertex(index, 2)];
		points = dof->type->points + dof->type->np_vert +
		    dof->type->np_edge * 2;
	    }
	    phgDofMapFaceData(dof->type, e, index, M);
	    for (j = 0; j < dof->type->np_face; j++) {
		sprintf(s, fmt, GlobalFace(g, i), j);
		if (dof->type->points != NULL) {
		    x = (*p0)[0] * points[0] + (*p1)[0] * points[1] +
			(*p2)[0] * points[2];
		    y = (*p0)[1] * points[0] + (*p1)[1] * points[1] +
			(*p2)[1] * points[2];
		    z = (*p0)[2] * points[0] + (*p1)[2] * points[1] +
			(*p2)[2] * points[2];
		    sprintf(s + strlen(s), "%s(%6.3lf,%6.3lf,%6.3lf) = ",
			    dof->name, (double)x, (double)y, (double)z);
		    points += 3;
		}
		for (k = 0; k < dof->dim; k++)
		    sprintf(s + strlen(s), "%10.4le ",
			    (double)*(data + dof->dim * M[j] + k));
		phgInfo(-1, "%s\n", s);
	    }
	    data += dof->count_face;
	}
    }

    if ((k = dof->type->np_elem) > 0) {
	fmt = (k <= 1) ? "%5d  " : ((k < 10) ? "%5d[%d]  " : "%5d[%02d]  ");
	phgInfo(-1, "Elements:\n");
	for (i = 0; i < g->nelem; i++) {
	    if (dof->type->points != NULL) {
		if ((e = tmp_elem[i].e) == NULL) {
		    data += dof->count_elem;
		    continue;
		}
		p0 = g->verts + e->verts[0];
		p1 = g->verts + e->verts[1];
		p2 = g->verts + e->verts[2];
		p3 = g->verts + e->verts[3];
		points = dof->type->points + dof->type->np_vert +
		    dof->type->np_edge * 2 + dof->type->np_face * 3;
	    }
	    for (j = 0; j < dof->type->np_elem; j++) {
		sprintf(s, fmt, GlobalElement(g, i), j);
		if (dof->type->points != NULL) {
		    x = (*p0)[0] * points[0] + (*p1)[0] * points[1] +
			(*p2)[0] * points[2] + (*p3)[0] * points[3];
		    y = (*p0)[1] * points[0] + (*p1)[1] * points[1] +
			(*p2)[1] * points[2] + (*p3)[1] * points[3];
		    z = (*p0)[2] * points[0] + (*p1)[2] * points[1] +
			(*p2)[2] * points[2] + (*p3)[2] * points[3];
		    sprintf(s + strlen(s), "%s(%6.3lf,%6.3lf,%6.3lf) = ",
			    dof->name, (double)x, (double)y, (double)z);
		    points += 4;
		}
		for (k = 0; k < dof->dim; k++)
		    sprintf(s + strlen(s), "%10.4le ", (double)*(data++));
		phgInfo(-1, "%s\n", s);
	    }
	}
    }

    phgFree(tmp_vert);
    tmp_vert = tmp_edge = tmp_face = tmp_elem = tmp_end = NULL;
}

static SIMPLEX *elem;
static FLOAT **pdata = NULL, **cdata = NULL, *cdata0[15], *cdata1[15];
static BOOLEAN cdata0_flag = FALSE, cdata1_flag = FALSE;
static int vmap0[NVert], emap0[NEdge], vmap1[NVert], emap1[NEdge];

static void
c2f_func0(DOF *dof, SIMPLEX *e0, int bno, const FLOAT *lambda, FLOAT *values)
{
    FLOAT lam[Dim + 1];

    Unused(bno);

    lam[0] = lambda[vmap0[0]] + (lam[1] = lambda[3] * 0.5);
    lam[2] = lambda[vmap0[2]];
    lam[3] = lambda[vmap0[3]];

    phgDofEval_(dof, elem, lam, NULL, 0, values, pdata);
}

static void
c2f_func1(DOF *dof, SIMPLEX *e1, int bno, const FLOAT *lambda, FLOAT *values)
{
    FLOAT lam[Dim + 1];

    Unused(bno);

    lam[1] = lambda[vmap1[1]] + (lam[0] = lambda[3] * 0.5);
    lam[2] = lambda[vmap1[2]];
    lam[3] = lambda[vmap1[3]];

    phgDofEval_(dof, elem, lam, NULL, 0, values, pdata);
}

static void
f2c_func(DOF *dof, SIMPLEX *e, int bno, const FLOAT *lambda, FLOAT *values)
{
    int *map;
    FLOAT lam[Dim + 1], **data;

    Unused(bno);

    if ((e->children[1] == NULL) ||
	(lambda[0] >= lambda[1] && e->children[0] != NULL)) {
	/* use child 0 */
	map = vmap0;
	lam[map[0]] = lambda[0] - lambda[1];
	lam[3] = 2. * lambda[1];
	lam[map[2]] = lambda[2];
	lam[map[3]] = lambda[3];
	e = e->children[0];
	data = cdata0;
	if (cdata0_flag == FALSE) {
	    cdata0_flag = TRUE;
	    /* VERTEX data */
	    data[3] = cdata[0];
	    data[map[0]] = pdata[0];
	    data[map[2]] = pdata[2];
	    data[map[3]] = pdata[3];
	    /* EDGE data */
	    data[NVert + GetEdgeNo(map[0], 3)] = cdata[1];
	    data[NVert + GetEdgeNo(map[2], 3)] = cdata[3];
	    data[NVert + GetEdgeNo(map[3], 3)] = cdata[4];
	    data[NVert + GetEdgeNo(map[0], map[2])] = pdata[NVert + 1];
	    data[NVert + GetEdgeNo(map[0], map[3])] = pdata[NVert + 2];
	    data[NVert + GetEdgeNo(map[2], map[3])] = pdata[NVert + 5];
	    /* FACE data */
	    data[NVert + NEdge + map[2]] = cdata[5];
	    data[NVert + NEdge + map[3]] = cdata[7];
	    data[NVert + NEdge + map[0]] = cdata[9];
	    data[NVert + NEdge + 3] = pdata[NVert + NEdge + 1];
	    /* ELEMENT data */
	    data[NVert + NEdge + NFace] = cdata[10];
	}
    }
    else {
	/* use child 1 */
	map = vmap1;
	lam[map[1]] = lambda[1] - lambda[0];
	lam[3] = 2. * lambda[0];
	lam[map[2]] = lambda[2];
	lam[map[3]] = lambda[3];
	e = e->children[1];
	data = cdata1;
	if (cdata1_flag == FALSE) {
	    cdata1_flag = TRUE;
	    /* VERTEX data */
	    data[3] = cdata[0];
	    data[map[1]] = pdata[1];
	    data[map[2]] = pdata[2];
	    data[map[3]] = pdata[3];
	    /* EDGE data */
	    data[NVert + GetEdgeNo(map[1], 3)] = cdata[2];
	    data[NVert + GetEdgeNo(map[2], 3)] = cdata[3];
	    data[NVert + GetEdgeNo(map[3], 3)] = cdata[4];
	    data[NVert + GetEdgeNo(map[1], map[2])] = pdata[NVert + 3];
	    data[NVert + GetEdgeNo(map[1], map[3])] = pdata[NVert + 4];
	    data[NVert + GetEdgeNo(map[2], map[3])] = pdata[NVert + 5];
	    /* FACE data */
	    data[NVert + NEdge + map[2]] = cdata[6];
	    data[NVert + NEdge + map[3]] = cdata[8];
	    data[NVert + NEdge + map[1]] = cdata[9];
	    data[NVert + NEdge + 3] = pdata[NVert + NEdge + 0];
	    /* ELEMENT data */
	    data[NVert + NEdge + NFace] = cdata[11];
	}
    }

    phgDofEval_(dof, e, lam, NULL, 0, values, data);
}

void
phgDofInterpC2FGeneric(DOF *dof, SIMPLEX *e, FLOAT **parent_data,
		       FLOAT **children_data)
/* Default coarse-to-fine interpolation function. It can be used
 * as the 'InterpC2F' member in any DOF_TYPE which has valid 'InitFunc'
 * and 'BasFuncs' members.
 *
 * Note:
 *	For better efficiency specialized functions should be implemented
 *	for specific DOF_TYPEs */
{
    SIMPLEX *e0 = e->children[0], *e1 = e->children[1];
    FLOAT **child0_data, **child1_data;

    phgMapP2C(e, vmap0, emap0, 0);
    phgMapP2C(e, vmap1, emap1, 1);

    /* Note: only need child[01]_data for HBn, HCn and HFEBn with n >= 2 */
    /* FIXME: do we need to add a flag in DOF_TYPE for this test? */
    if (dof->type->order > 1 /* other tests */) {
	/* setup complete list of pointers for children data */
	child0_data = phgAlloc(2 * 15 * sizeof(*child0_data));
	child1_data = child0_data + 15;

	/* the vertices of child0 */
	child0_data[vmap0[0]] = parent_data[0];
	child0_data[vmap0[2]] = parent_data[2];
	child0_data[vmap0[3]] = parent_data[3];
	child0_data[3] = children_data[0];

	/* the vertices of child1 */
	child1_data[vmap1[1]] = parent_data[1];
	child1_data[vmap1[2]] = parent_data[2];
	child1_data[vmap1[3]] = parent_data[3];
	child1_data[3] = children_data[0];

	/* the edges of child0 */
	child0_data[4 + emap0[1]] = parent_data[5];	/* parent edge 0-2 */
	child0_data[4 + emap0[2]] = parent_data[6];	/* parent edge 0-3 */
	child0_data[4 + emap0[5]] = parent_data[9];	/* parent edge 2-3 */
	/* cut edge (v0 mapped to 2 or 0) */
	if (vmap0[0] == 2)
	    child0_data[4 + 5] = children_data[1];	/* cut edge (2-3) */
	else
	    child0_data[4 + 2] = children_data[1];	/* cut edge (0-3) */
	/* new edge on v2 (v2 mapped to 1) */
	child0_data[4 + 4] = children_data[3];		/* new edge (1-3) */
	/* new edge on v3 (v3 mapped to 0 or 2) */
	if (vmap0[3] == 0)
	    child0_data[4 + 2] = children_data[4];	/* new edge (0-3) */
	else
	    child0_data[4 + 5] = children_data[4];	/* new edge (2-3) */

	/* the edges of child1 */
	child1_data[4 + emap1[3]] = parent_data[7];	/* parent edge 1-2 */
	child1_data[4 + emap1[4]] = parent_data[8];	/* parent edge 1-3 */
	child1_data[4 + emap1[5]] = parent_data[9];	/* parent edge 2-3 */
	/* cut edge  (v1 mapped to 2 or 0) */
	if (vmap1[1] == 2)
	    child1_data[4 + 5] = children_data[2];	/* cut edge (2-3) */
	else
	    child1_data[4 + 2] = children_data[2];	/* cut edge (0-3) */
	/* new edge on v2 (v2 mapped to 2 or 1) */
	if (vmap1[2] == 2)
	    child1_data[4 + 5] = children_data[3];	/* new edge (2-3) */
	else
	    child1_data[4 + 4] = children_data[3];	/* new edge (1-3) */
	/* new edge on v3 (v3 mapped to 0, 1 or 2) */
	if (vmap1[3] == 0)
	    child1_data[4 + 2] = children_data[4];	/* new edge (0-3) */
	else if (vmap1[3] == 1)
	    child1_data[4 + 4] = children_data[4];	/* new edge (1-3) */
	else
	    child1_data[4 + 5] = children_data[4];	/* new edge (2-3) */

	/* the faces of child0 */
	child0_data[10 + 3] = parent_data[11];		/* parent face 1 */
	child0_data[10 + vmap0[0]] = children_data[9];	/* new face */
	child0_data[10 + vmap0[2]] = children_data[5];	/* cut face */
	child0_data[10 + vmap0[3]] = children_data[7];	/* cut face */

	/* the faces of child1 */
	child1_data[10 + 3] = parent_data[10];		/* parent face 0 */
	child1_data[10 + vmap1[1]] = children_data[9];	/* new face */
	child1_data[10 + vmap1[2]] = children_data[6];	/* cut face */
	child1_data[10 + vmap1[3]] = children_data[8];	/* cut face */

	/* the volume data */
	child0_data[14] = children_data[10];
	child1_data[14] = children_data[11];
    }
    else {
	child0_data = child1_data = NULL;
    }

    elem = e;
    pdata = parent_data;

    if (dof->type->np_vert > 0) {
	/* The new vertex */
	dof->type->InitFunc(dof, e0, VERTEX, 3, NULL,
			    c2f_func0, NULL, children_data[0], child0_data);
    }

    if (dof->type->np_edge > 0) {
	/* The cut-edges */
	dof->type->InitFunc(dof, e0, EDGE, emap0[0], NULL,
			    c2f_func0, NULL, children_data[1], child0_data);
	dof->type->InitFunc(dof, e1, EDGE, emap1[0], NULL,
			    c2f_func1, NULL, children_data[2], child1_data);
	/* The new edges */
	dof->type->InitFunc(dof, e0, EDGE, GetEdgeNo(vmap0[2], 3), NULL,
			    c2f_func0, NULL, children_data[3], child0_data);
	dof->type->InitFunc(dof, e0, EDGE, GetEdgeNo(vmap0[3], 3), NULL,
			    c2f_func0, NULL, children_data[4], child0_data);
    }

    if (dof->type->np_face > 0) {
	/* the cut faces of e0 */
	dof->type->InitFunc(dof, e0, FACE, vmap0[2], NULL,
			    c2f_func0, NULL, children_data[5], child0_data);
	dof->type->InitFunc(dof, e0, FACE, vmap0[3], NULL,
			    c2f_func0, NULL, children_data[7], child0_data);
	/* the cut faces of e1 */
	dof->type->InitFunc(dof, e1, FACE, vmap1[2], NULL,
			    c2f_func1, NULL, children_data[6], child1_data);
	dof->type->InitFunc(dof, e1, FACE, vmap1[3], NULL,
			    c2f_func1, NULL, children_data[8], child1_data);
	/* the new face */
	dof->type->InitFunc(dof, e0, FACE, vmap0[0], NULL,
			    c2f_func0, NULL, children_data[9], child0_data);
    }

    if (dof->type->np_elem > 0) {
	/* volume data on e0 */
	dof->type->InitFunc(dof, e0, ELEMENT, 0, NULL,
			    c2f_func0, NULL, children_data[10], child0_data);
	/* volume data on e1 */
	dof->type->InitFunc(dof, e1, ELEMENT, 0, NULL,
			    c2f_func1, NULL, children_data[11], child1_data);
    }

    phgFree(child0_data);
}

void
phgDofInterpF2CGeneric(DOF *dof, SIMPLEX *e, FLOAT **parent_data,
		       FLOAT **children_data)
/* Default fine-to-coarse interpolation function. It can be used
 * as the 'InterpF2C' member in any DOF_TYPE which has valid 'InitFunc'
 * and 'BasFuncs' members.
 *
 * Note:
 *   1. It is assumed that 'parent_data' already contains valid values at
 *	the locations common to parent and children, only values at edge 0,
 *	faces 2 and 3, and element data need to be computed.
 *
 *   2. For better efficiency specialized functions should be implemented
 *	for specific DOF_TYPEs */
{

    pdata = parent_data;
    cdata = children_data;
    cdata0_flag = FALSE;
    cdata1_flag = FALSE;

    phgMapP2C(e, vmap0, emap0, 0);
    phgMapP2C(e, vmap1, emap1, 1);

    if (dof->type->np_edge > 0) {
	dof->type->InitFunc(dof, e, EDGE, 0, NULL, f2c_func, NULL,
			    parent_data[NVert + 0], parent_data);
    }

    if (dof->type->np_face > 0) {
	dof->type->InitFunc(dof, e, FACE, 2, NULL, f2c_func, NULL,
			    parent_data[NVert + NEdge + 2], parent_data);
	dof->type->InitFunc(dof, e, FACE, 3, NULL, f2c_func, NULL,
			    parent_data[NVert + NEdge + 3], parent_data);
    }

    if (dof->type->np_elem > 0) {
	dof->type->InitFunc(dof, e, ELEMENT, 0, NULL, f2c_func, NULL,
			    parent_data[NVert + NEdge + NFace], parent_data);
    }
}

void
phgDofInitFuncPoint(DOF *dof, SIMPLEX *e, GTYPE type, int index,
		    DOF_USER_FUNC userfunc,
		    DOF_USER_FUNC_LAMBDA userfunc_lambda,
		    const FLOAT *funcvalues, FLOAT *dofvalues,
		    FLOAT **pdofvalues)
{
    const GRID *g = dof->g;
    COORD *p0 = NULL, *p1 = NULL, *p2 = NULL, *p3;
    FLOAT *points;
    FLOAT x, y, z, lambda[Dim + 1];
    int n, j, v0, v1, v2, bno = 0;

    Unused(pdofvalues);
    assert(dof->type->dim == 1);	/* only 1 is meaningful */

    switch (type) {
	case VERTEX:
	    n = dof->type->np_vert;
	    break;
	case EDGE:
	    n = dof->type->np_edge;
	    break;
	case FACE:
	    n = dof->type->np_face;
	    break;
	case ELEMENT:
	    n = dof->type->np_elem;
	    break;
	default:
	    n = 0;		/* make gcc happy */
    }

    if (funcvalues != NULL) {
	memcpy(dofvalues, funcvalues, n * dof->dim * sizeof(*dofvalues));
	return;
    }

    if (dof->type->points == NULL)
	phgError(1, "%s:%d, not a point DOF.\n", __FILE__, __LINE__);

    if (userfunc == NULL && userfunc_lambda == NULL)
	phgError(1, "%s:%d, invalid user function!\n", __FILE__, __LINE__);

    switch (type) {
	case VERTEX:
	    if (userfunc_lambda != NULL) {
		lambda[0] = lambda[1] = lambda[2] = lambda[3] = 0.;
		lambda[index] = 1.0;
		bno = index * n;
		for (j = 0; j < n; j++) {
		    userfunc_lambda(dof, e, bno + j, lambda, dofvalues);
		    dofvalues += dof->dim;
		}
		break;
	    }
	    x = g->verts[e->verts[index]][0];
	    y = g->verts[e->verts[index]][1];
	    z = g->verts[e->verts[index]][2];
	    for (j = 0; j < n; j++) {
		userfunc(x, y, z, dofvalues);
		dofvalues += dof->dim;
	    }
	    break;
	case EDGE:
	    v0 = GetEdgeVertex(index, 0);
	    v1 = GetEdgeVertex(index, 1);
	    if (userfunc_lambda != NULL) {
		lambda[0] = lambda[1] = lambda[2] = lambda[3] = 0.;
		bno = NVert * dof->type->np_vert + index * n;
	    }
	    else {
		p0 = g->verts + e->verts[v0];
		p1 = g->verts + e->verts[v1];
	    }
	    points = dof->type->points + dof->type->np_vert;
	    if (n > 1) {
		int M[n];
		phgDofMapEdgeData(dof->type, e, index, M);
		if (userfunc_lambda == NULL) {
		    for (j = 0; j < n; j++) {
			x = (*p0)[0] * points[0] + (*p1)[0] * points[1];
			y = (*p0)[1] * points[0] + (*p1)[1] * points[1];
			z = (*p0)[2] * points[0] + (*p1)[2] * points[1];
			userfunc(x, y, z, dofvalues + M[j] * dof->dim);
			points += 2;
		    }
		}
		else {
		    for (j = 0; j < n; j++) {
			lambda[v0] = points[0];
			lambda[v1] = points[1];
			userfunc_lambda(dof, e, bno + j, lambda,
					dofvalues + M[j] * dof->dim);
			points += 2;
		    }
		}
	    }
	    else {
		if (userfunc_lambda == NULL) {
		    for (j = 0; j < n; j++) {
			x = (*p0)[0] * points[0] + (*p1)[0] * points[1];
			y = (*p0)[1] * points[0] + (*p1)[1] * points[1];
			z = (*p0)[2] * points[0] + (*p1)[2] * points[1];
			userfunc(x, y, z, dofvalues);
			points += 2;
			dofvalues += dof->dim;
		    }
		}
		else {
		    for (j = 0; j < n; j++) {
			lambda[v0] = points[0];
			lambda[v1] = points[1];
			userfunc_lambda(dof, e, bno + j, lambda, dofvalues);
			points += 2;
			dofvalues += dof->dim;
		    }
		}
	    }
	    break;
	case FACE:
	    v0 = GetFaceVertex(index, 0);
	    v1 = GetFaceVertex(index, 1);
	    v2 = GetFaceVertex(index, 2);
	    if (userfunc_lambda != NULL) {
		lambda[0] = lambda[1] = lambda[2] = lambda[3] = 0.;
		bno = NVert * dof->type->np_vert + NEdge * dof->type->np_edge +
		      index * n;
	    }
	    else {
		p0 = g->verts + e->verts[v0];
		p1 = g->verts + e->verts[v1];
		p2 = g->verts + e->verts[v2];
	    }
	    points = dof->type->points + dof->type->np_vert +
		dof->type->np_edge * 2;
	    if (n > 1) {
		int M[n];
		phgDofMapFaceData(dof->type, e, index, M);
		if (userfunc_lambda == NULL) {
		    for (j = 0; j < n; j++) {
			x = (*p0)[0] * points[0] + (*p1)[0] * points[1] +
			    (*p2)[0] * points[2];
			y = (*p0)[1] * points[0] + (*p1)[1] * points[1] +
			    (*p2)[1] * points[2];
			z = (*p0)[2] * points[0] + (*p1)[2] * points[1] +
			    (*p2)[2] * points[2];
			userfunc(x, y, z, dofvalues + M[j] * dof->dim);
			points += 3;
		    }
		}
		else {
		    for (j = 0; j < n; j++) {
			lambda[v0] = points[0];
			lambda[v1] = points[1];
			lambda[v2] = points[2];
			userfunc_lambda(dof, e, bno + j, lambda,
					dofvalues + M[j] * dof->dim);
			points += 3;
		    }
		}
	    }
	    else {
		if (userfunc_lambda == NULL) {
		    for (j = 0; j < n; j++) {
			x = (*p0)[0] * points[0] + (*p1)[0] * points[1] +
			    (*p2)[0] * points[2];
			y = (*p0)[1] * points[0] + (*p1)[1] * points[1] +
			    (*p2)[1] * points[2];
			z = (*p0)[2] * points[0] + (*p1)[2] * points[1] +
			    (*p2)[2] * points[2];
			userfunc(x, y, z, dofvalues);
			points += 3;
			dofvalues += dof->dim;
		    }
		}
		else {
		    for (j = 0; j < n; j++) {
			lambda[v0] = points[0];
			lambda[v1] = points[1];
			lambda[v2] = points[2];
			userfunc_lambda(dof, e, bno + j, lambda, dofvalues);
			points += 3;
			dofvalues += dof->dim;
		    }
		}
	    }
	    break;
	case ELEMENT:
	    points = dof->type->points + dof->type->np_vert +
		dof->type->np_edge * 2 + dof->type->np_face * 3;
	    if (userfunc_lambda == NULL) {
		p0 = g->verts + e->verts[0];
		p1 = g->verts + e->verts[1];
		p2 = g->verts + e->verts[2];
		p3 = g->verts + e->verts[3];
		for (j = 0; j < n; j++) {
		    x = (*p0)[0] * points[0] + (*p1)[0] * points[1] +
			(*p2)[0] * points[2] + (*p3)[0] * points[3];
		    y = (*p0)[1] * points[0] + (*p1)[1] * points[1] +
			(*p2)[1] * points[2] + (*p3)[1] * points[3];
		    z = (*p0)[2] * points[0] + (*p1)[2] * points[1] +
			(*p2)[2] * points[2] + (*p3)[2] * points[3];
		    userfunc(x, y, z, dofvalues);
		    dofvalues += dof->dim;
		    points += 4;
		}
	    }
	    else {
		bno = NVert * dof->type->np_vert + NEdge * dof->type->np_edge +
		      NFace * dof->type->np_face + index * n;
		for (j = 0; j < n; j++) {
		    userfunc_lambda(dof, e, bno + j, points, dofvalues);
		    dofvalues += dof->dim;
		    points += 4;
		}
	    }
	    break;
    }
}

void
phgDofSetDataByFunction_(DOF *dof, DOF_USER_FUNC userfunc,
			 DOF_USER_FUNC_LAMBDA userfunc_lambda,
			 FLOAT *funcvalues)
{
    GRID *g = dof->g;
    FLOAT *data;
    INT i;
    size_t size;

    if (dof->type == DOF_ANALYTIC)
	return;

    size = (dof->type == DOF_CONSTANT) ? dof->dim : DofGetDataCount(dof);
    if (dof->data == NULL) {
	dof->data = phgCalloc(size, sizeof(*dof->data));
	dof->data_vert = dof->data;
	dof->data_edge = dof->data_vert + DofGetVertexDataCount(dof);
	dof->data_face = dof->data_edge + DofGetEdgeDataCount(dof);
	dof->data_elem = dof->data_face + DofGetFaceDataCount(dof);
    }

    if (userfunc == DofNoAction || userfunc == DofInterpolation)
	return;

    if (dof->type == DOF_CONSTANT) {
	/* constant DOF, using funcvalues */
	if (funcvalues != NULL)
	    memcpy(dof->data, funcvalues, dof->dim * sizeof(*dof->data));
	else if (userfunc != NULL)
	    userfunc(0., 0., 0., dof->data);
	else {
	    FLOAT lambda[] = { 0., 0., 0., 1. };
	    assert(userfunc_lambda != NULL);
	    userfunc_lambda(dof, dof->g->roots, -1, lambda, dof->data);
	}
	return;
    }

    if ((userfunc != NULL || userfunc_lambda != NULL)
	&& dof->type->InitFunc == NULL)
	phgError(1,
		 "phgDofSetDataByFunction_: no initialization function!\n");

    data = dof->data;
    setup_tmp(dof);

    if (dof->type->np_vert > 0) {
	size = dof->count_vert;
	for (i = 0; i < g->nvert; i++) {
	    if (tmp_vert[i].flag
		&& (!flagged_only || tmp_vert[i].e->flag < 1))
		dof->type->InitFunc(dof, tmp_vert[i].e, VERTEX,
				    tmp_vert[i].index, userfunc,
				    userfunc_lambda, funcvalues, data, NULL);
	    data += size;
	}
    }

    if (dof->type->np_edge > 0) {
	size = dof->count_edge;
	for (i = 0; i < g->nedge; i++) {
	    if (tmp_edge[i].flag
		&& (!flagged_only || tmp_edge[i].e->flag < 1))
		dof->type->InitFunc(dof, tmp_edge[i].e, EDGE,
				    tmp_edge[i].index, userfunc,
				    userfunc_lambda, funcvalues, data, NULL);
	    data += size;
	}
    }

    if (dof->type->np_face > 0) {
	size = dof->count_face;
	for (i = 0; i < g->nface; i++) {
	    if (tmp_face[i].flag
		&& (!flagged_only || tmp_face[i].e->flag < 1))
		dof->type->InitFunc(dof, tmp_face[i].e, FACE,
				    tmp_face[i].index, userfunc,
				    userfunc_lambda, funcvalues, data, NULL);
	    data += size;
	}
    }

    if (dof->type->np_elem > 0) {
	size = dof->count_elem;
	for (i = 0; i < g->nelem; i++) {
	    if (tmp_elem[i].flag
		&& (!flagged_only || tmp_elem[i].e->flag < 1))
		dof->type->InitFunc(dof, tmp_elem[i].e, ELEMENT,
				    tmp_elem[i].index, userfunc,
				    userfunc_lambda, funcvalues, data, NULL);
	    data += size;
	}
    }

    phgFree(tmp_vert);
    tmp_vert = tmp_edge = tmp_face = tmp_elem = tmp_end = NULL;
}

static int dummy_n;
static const FLOAT *dummy_values;
static void
dummy_userfunc_lambda(DOF *dof, SIMPLEX *e, int bno, const FLOAT *lambda,
			FLOAT *values)
{
    int i, n;
    const FLOAT *p;

    Unused(e);
    Unused(lambda);
    Unused(bno);

    n = DofDim(dof);
    for (i = 0, p = dummy_values; i < n; i++) {
	*(values++) = *(p++);
	if (p >= dummy_values + dummy_n)
	    p = dummy_values;
    }
}

void
phgDofSetDataByValue(DOF *dof, FLOAT value)
{
    size_t i, np, nvalues;
    FLOAT *values;

    assert(dof->type != DOF_ANALYTIC);

    if (dof->type != DOF_CONSTANT) {
	nvalues = DofDim(dof);
	if (dof->type->InitFunc != phgDofInitFuncPoint) {
	    if (dof->type->InitFunc == NULL) {
		/* no init function, simply copy the data */
		for (i = 0; i < DofGetDataCount(dof); i++)
		    dof->data[i] = value;
		return;
	    }
	    dummy_n = 1;
	    dummy_values = &value;
	    phgDofSetDataByFunction_(dof, NULL, dummy_userfunc_lambda, NULL);
	    return;
	}
	np = dof->type->np_vert;
	if (np < (i = dof->type->np_edge))
	    np = i;
	if (np < (i = dof->type->np_face))
	    np = i;
	if (np < (i = dof->type->np_elem))
	    np = i;
    }
    else {
	nvalues = dof->dim;
	np = 1;
    }
    values = phgAlloc(nvalues * np * sizeof(*values));
    /* make 'np' * 'nvalues' duplicate copies of the value */
    for (i = 0; i < np * nvalues; i++)
	values[i] = value;
    phgDofSetDataByFunction_(dof, NULL, NULL, values);
    phgFree(values);
}

void
phgDofSetDataByValues(DOF *dof, const FLOAT array_of_values[])
{
    size_t i, np, nvalues;
    FLOAT *values;

    assert(dof->type != DOF_ANALYTIC);

    if (dof->type != DOF_CONSTANT) {
	nvalues = DofDim(dof);
	if (dof->type->InitFunc != phgDofInitFuncPoint) {
	    dummy_n = nvalues;
	    dummy_values = array_of_values;
	    phgDofSetDataByFunction_(dof, NULL, dummy_userfunc_lambda, NULL);
	    return;
	}
	np = dof->type->np_vert;
	if (np < (i = dof->type->np_edge))
	    np = i;
	if (np < (i = dof->type->np_face))
	    np = i;
	if (np < (i = dof->type->np_elem))
	    np = i;
    }
    else {
	nvalues = dof->dim;
	np = 1;
    }
    values = phgAlloc(nvalues * np * sizeof(*values));
    /* make 'np' duplicate copies of the values */
    for (i = 0; i < nvalues; i++)
	values[i] = (i < nvalues) ? array_of_values[i] : values[i - nvalues];
    phgDofSetDataByFunction_(dof, NULL, NULL, values);
    phgFree(values);
}

void
phgDofSetDataByValuesV(DOF *dof, FLOAT v0, ...)
/* set DOF data using variable arguments */
{
    int i, n = DofDim(dof);
    FLOAT *values;
    va_list ap;                                                         \

    values = phgAlloc(n * sizeof(*values));
    values[0] = v0;
    va_start(ap, v0);
    for (i = 1; i < n; i++)
	values[i] = va_arg(ap, FLOAT);
    va_end(ap);
    phgDofSetDataByValues(dof, values);
}

void
phgDofScatter(GRID *g)
/* scatters DOFs to slaves (called by phgPartitionGrid) */
{

    return;
}

static int level = 0;		/* recursion level */
/* common values and pointers used by the interpolation routines */
static int n0, n1, n2, n3;
static FLOAT *data0, *data1, *data2, *data3;

void
update_dof_data(DOF *dof, SIMPLEX *e, FLOAT **data)
/* places DOF data for element 'e' pointed by 'data' to dof->data */
{
    int i, j;
    int size[4];
    int count[] = { NVert, NEdge, NFace, 1 };
    int *index[4];
    FLOAT *loc[4];
    FLOAT *p;

    size[0] = n0;
    size[1] = n1;
    size[2] = n2;
    size[3] = n3;
    index[0] = e->verts;
    index[1] = e->edges;
    index[2] = e->faces;
    index[3] = &e->index;
    loc[0] = data0;
    loc[1] = data1;
    loc[2] = data2;
    loc[3] = data3;

    for (j = 0; j < 4; j++) {
	for (i = 0; i < count[j]; i++) {
	    p = loc[j] + index[j][i] * size[j];
	    if (size[j] > 0 && p != *data) {
		memcpy(p, *data, size[j] * sizeof(FLOAT));
	    }
	    data++;
	}
    }
}

static void
build_child_pointers(DOF *dof, SIMPLEX *e, FLOAT **old_data, FLOAT **new_data,
		     FLOAT *buffer, int child)
/* construct list of pointers for children */
{
    int i, Vmap[NVert], Emap[NEdge];
    FLOAT *p1, *p2, *p3;

    phgMapC2P(e, Vmap, Emap, child);

    p1 = buffer;
    p2 = p1 + n1 * 4;		/* 4 new edges */
    p3 = p2 + n2 * 5;		/* 5 new faces */

    /* vertices */
    if (n0 > 0) {
	for (i = 0; i < NVert - 1; i++) {
	    *(new_data++) = old_data[Vmap[i]];
	}
	*(new_data++) = data0 + e->children[0]->verts[NVert - 1] * n0;
    }
    else {
	new_data += NVert;
    }

    /* edges */
    if (n1 > 0) {
	for (i = 0; i < NEdge; i++) {
	    switch (Emap[i]) {
		case -1:	/* new edge */
		    (*new_data++) = p1 + (Vmap[GetEdgeVertex(i, 0)]) * n1;
		    break;
		case 0:	/* cut edge */
		    (*new_data++) = p1 + child * n1;
		    break;
		default:	/* old edge */
		    (*new_data++) = old_data[NVert + Emap[i]];
		    break;
	    }
	}
    }
    else {
	new_data += NEdge;
    }

    /* faces */
    if (n2 > 0) {
	for (i = 0; i < NFace; i++) {
	    switch (Vmap[i]) {
		case 0:
		case 1:	/* new face */
		    (*new_data++) = p2 + 4 * n2;
		    break;
		case 2:	/* cut face on F2 */
		    (*new_data++) = p2 + child * n2;
		    break;
		case 3:	/* cut face on F3 */
		    (*new_data++) = p2 + (child + 2) * n2;
		    break;
		case -1:	/* old faces */
		    (*new_data++) = old_data[NVert + NEdge + 1 - child];
		    break;
	    }
	}
    }
    else {
	new_data += NFace;
    }

    /* volume */
    if (n3 > 0) {
	(*new_data) = p3 + child * n3;
    }
}

static void
interp_recursive(DOF *dof, SIMPLEX *e, FLOAT **old_data)
/* recursively interpolates DOFs down the binary tree starting from e */
{
    static FLOAT *p1, *p2, *p3;
    static SIMPLEX *e0, *e1;
    FLOAT *buffer, *new_data[15];

    if ((e0 = e->children[0]) == NULL || (e1 = e->children[1]) == NULL) {
	/* copy data to dof->data */
	update_dof_data(dof, e, old_data);
	return;
    }

    level++;

    /* allocate buffer for new DOFs (only allocate for edge/face/vol data) */
    if (n1 > 0 || n2 > 0 || n3 > 0)
	buffer = phgAlloc((n1 * 4 + n2 * 5 + n3 * 2) * sizeof(*buffer));
    else
	buffer = NULL;

    p1 = buffer;
    p2 = p1 + n1 * 4;
    p3 = p2 + n2 * 5;

    memset(new_data, 0, sizeof(new_data));

    if (dof->type->InterpC2F != NULL && dof->userfunc == DofInterpolation) {
	/* interpolation */
	if (n0 > 0) {
	    /* the new vertex */
	    new_data[0] = data0 + e0->verts[3] * n0;
	}

	if (n1 > 0) {
	    /* the cut edge */
	    new_data[1] = p1;
	    new_data[2] = p1 + n1;
	    /* the new edge on V2 */
	    new_data[3] = p1 + n1 * 2;
	    /* the new edge on V3 */
	    new_data[4] = p1 + n1 * 3;
	}

	if (n2 > 0) {
	    /* the cut faces on F2 */
	    new_data[5] = p2;
	    new_data[6] = p2 + n2;
	    /* the cut faces on F3 */
	    new_data[7] = p2 + n2 * 2;
	    new_data[8] = p2 + n2 * 3;
	    /* the new face */
	    new_data[9] = p2 + n2 * 4;
	}

	if (n3 > 0) {
	    /* volume data for child 0 */
	    new_data[10] = p3;
	    /* volume data for child 1 */
	    new_data[11] = p3 + n3;
	}

	dof->type->InterpC2F(dof, e, old_data, new_data);
    }

    if (e0 != NULL) {
	build_child_pointers(dof, e, old_data, new_data, buffer, 0);
	interp_recursive(dof, e0, new_data);
	e1 = e->children[1];	/* note: e1 changed by recursion */
    }

    if (e1 != NULL) {
	build_child_pointers(dof, e, old_data, new_data, buffer, 1);
	interp_recursive(dof, e1, new_data);
    }

    phgFree(buffer);

    level--;
}

void
phgDofInterpolate(GRID *g, BOOLEAN flag)
/* interpolates DOFs, should be called with flag==FALSE
   before refinement, then with flag==TRUE after refinement. */
{
    static SIMPLEX **elems = NULL, **pe, *e;
    static INT count, nvert, nedge, nface, nelem, nleaf;
    DOF **dof;

    /* check whether interpolation is really needed */
    if (g->dof == NULL)
	return;
    for (dof = g->dof; *dof != NULL; dof++) {
	if (!SpecialDofType((*dof)->type) && (*dof)->data != NULL)
	    break;
    }
    if (*dof == NULL)
	return;

    if ((!flag && elems != NULL) || (flag && elems == NULL))
	phgError(1, "phgDofInterpolate: incorrect calling sequence.\n");

    if (!flag) {
	SIMPLEX *e;
	/* store current leaf elements */
	pe = elems = phgAlloc((count = g->nleaf) * sizeof(*elems));
	ForAllElements(g, e)
	    * (pe++) = e;
	assert(pe - elems == count);
	/* save current DOF counts */
	nvert = g->nvert;
	nedge = g->nedge;
	nface = g->nface;
	nelem = g->nelem;
	nleaf = g->nleaf;

	return;
    }

    /* reset cache, map, etc. */
    /* phgDofClearCache(NULL, FALSE); (not necessary: elements not changed) */

    if (g->nleaf == nleaf)
	goto end;		/* no new elements created */

    /* Clear e->flag */
    ForAllElements(g, e)
	e->flag = 0;

    /* loop on DOFs */
    for (dof = g->dof; *dof != NULL; dof++) {
	size_t size, old_size;
	FLOAT *old_data, *p0, *p1, *p2, *p3;

	if (SpecialDofType((*dof)->type) || (*dof)->data == NULL)
	    continue;

	/* allocate new buffer */
	size = DofGetDataCount(*dof);

	if ((*dof)->userfunc == DofNoAction) {
	    phgFree((*dof)->data);
	    (*dof)->data = phgCalloc(size, sizeof(*(*dof)->data));
	    (*dof)->data_vert = (*dof)->data;
	    (*dof)->data_edge = (*dof)->data_vert +
		DofGetVertexDataCount(*dof);
	    (*dof)->data_face = (*dof)->data_edge + DofGetEdgeDataCount(*dof);
	    (*dof)->data_elem = (*dof)->data_face + DofGetFaceDataCount(*dof);
	    continue;
	}

	old_data = (*dof)->data;
	(*dof)->data = phgCalloc(size, sizeof(*(*dof)->data));
	(*dof)->data_vert = (*dof)->data;
	(*dof)->data_edge = (*dof)->data_vert + DofGetVertexDataCount(*dof);
	(*dof)->data_face = (*dof)->data_edge + DofGetEdgeDataCount(*dof);
	(*dof)->data_elem = (*dof)->data_face + DofGetFaceDataCount(*dof);
	/* copy old data to new buffer, because some entries may be only
	   referenced by non-leaf elements */
	p0 = (*dof)->data;
	p1 = old_data;
	if ((*dof)->count_vert > 0) {
	    old_size = (*dof)->count_vert * nvert;
	    memcpy(p0, p1, old_size * sizeof(*old_data));
	    p0 += (*dof)->count_vert * g->nvert;
	    p1 += old_size;
	}
	if ((*dof)->count_edge > 0) {
	    old_size = (*dof)->count_edge * nedge;
	    memcpy(p0, p1, old_size * sizeof(*old_data));
	    p0 += (*dof)->count_edge * g->nedge;
	    p1 += old_size;
	}
	if ((*dof)->count_face > 0) {
	    old_size = (*dof)->count_face * nface;
	    memcpy(p0, p1, old_size * sizeof(*old_data));
	    p0 += (*dof)->count_face * g->nface;
	    p1 += old_size;
	}
	if ((*dof)->count_elem > 0) {
	    old_size = (*dof)->count_elem * nelem;
	    memcpy(p0, p1, old_size * sizeof(*old_data));
	    p0 += (*dof)->count_elem * g->nelem;
	}

	n0 = (*dof)->count_vert;
	n1 = (*dof)->count_edge;
	n2 = (*dof)->count_face;
	n3 = (*dof)->count_elem;

	data0 = (*dof)->data;
	data1 = data0 + n0 * g->nvert;
	data2 = data1 + n1 * g->nedge;
	data3 = data2 + n2 * g->nface;

	p0 = old_data;
	p1 = p0 + n0 * nvert;
	p2 = p1 + n1 * nedge;
	p3 = p2 + n2 * nface;

	for (pe = elems; pe < elems + count; pe++) {
	    FLOAT *old[15];
	    if (n0 > 0) {
		old[0] = p0 + (*pe)->verts[0] * n0;
		old[1] = p0 + (*pe)->verts[1] * n0;
		old[2] = p0 + (*pe)->verts[2] * n0;
		old[3] = p0 + (*pe)->verts[3] * n0;
	    }
	    if (n1 > 0) {
		old[4] = p1 + (*pe)->edges[0] * n1;
		old[5] = p1 + (*pe)->edges[1] * n1;
		old[6] = p1 + (*pe)->edges[2] * n1;
		old[7] = p1 + (*pe)->edges[3] * n1;
		old[8] = p1 + (*pe)->edges[4] * n1;
		old[9] = p1 + (*pe)->edges[5] * n1;
	    }
	    if (n2 > 0) {
		old[10] = p2 + (*pe)->faces[0] * n2;
		old[11] = p2 + (*pe)->faces[1] * n2;
		old[12] = p2 + (*pe)->faces[2] * n2;
		old[13] = p2 + (*pe)->faces[3] * n2;
	    }
	    if (n3 > 0) {
		old[14] = p3 + (*pe)->index * n3;
	    }
	    (*pe)->flag = 1;	/* element is on old mesh */
	    interp_recursive(*dof, *pe, old);
	    /* TODO: make use of flags[] to avoid redundant computations */
	}
	phgFree(old_data);
	flagged_only = TRUE;
	phgDofSetDataByFunction(*dof, (*dof)->userfunc);
	flagged_only = FALSE;
    }

  end:
    /* Free memory */
    phgFree(elems);
    elems = NULL;
}

#define IGNORE_NOACTION	0

void
phgDofRedistribute(GRID *g, BOOLEAN flag)
/* redistributes DOFs, should be called with flag==FALSE
   before the grid is changed, then with flag==TRUE after.

   Since the size of migrated data is normally small, we use
   a circular algorithm in which the lists of missing data
   are circulated on the processes ring and filled up on the fly

   TODO: determine which part of data is available on which process by
	 circulating the missing indices, then use MPI_Alltoallv
	 to exchange actual data (less communications when many DOFs) */
{
    Unused(g);
    Unused(flag);

    return;
}

/*----------------- utilities ---------------*/

INT
phgDofMapE2D(DOF *u, SIMPLEX *e, int index)
/* returns DOF index of the unknown with element index 'index' on 'e' */
{
    GRID *g = u->g;
    int m, n;
    INT index0;

    if (index < (n = (m = u->count_vert) * NVert)) {
	return m * e->verts[index / m] + (index % m);
    }
    index -= n;
    index0 = m * g->nvert;

    if (index < (n = (m = u->count_edge) * NEdge)) {
	if (u->type->np_edge > 1) {
	    int k = index / m, M[u->type->np_edge];
	    phgDofMapEdgeData(u->type, e, k, M);
	    index %= m;
	    return index0 + m * e->edges[k] + M[index / u->dim] * u->dim +
			index % u->dim;
	}
	else {
	    return index0 + m * e->edges[index / m] + (index % m);
	}
    }
    index -= n;
    index0 += m * g->nedge;

    if (index < (n = (m = u->count_face) * NFace)) {
	if (u->type->np_face > 1) {
	    int k = index / m, M[u->type->np_face];
	    phgDofMapFaceData(u->type, e, k, M);
	    index %= m;
	    return index0 + m * e->faces[k] + M[index / u->dim] * u->dim +
			index % u->dim;
	}
	else {
	    return index0 + m * e->faces[index / m] + (index % m);
	}
    }
    index -= n;
    index0 += m * g->nface;

    return index0 + u->count_elem * e->index + index;
}

BTYPE
phgDofGetBoundaryType(DOF *u, INT index)
/* returns boundary type of the unknown with local index 'index' */
{
    INT n;

    if (index < (n = DofGetVertexDataCount(u)))
	return u->g->types_vert[index / u->count_vert];
    index -= n;

    if (index < (n = DofGetEdgeDataCount(u)))
	return u->g->types_edge[index / u->count_edge];
    index -= n;

    if (index < (n = DofGetFaceDataCount(u)))
	return u->g->types_face[index / u->count_face];
    index -= n;

    assert(index < DofGetElementDataCount(u));
    return u->g->types_elem[index / u->count_elem];
}

BTYPE
phgDofGetElementBoundaryType(DOF *u, SIMPLEX *e, int index)
/* returns boundary type of the unknown with local 'index' on element 'e' */
{
    GRID *g = u->g;
    int m, n;

    if (index < (n = (m = u->count_vert) * NVert))
	return g->types_vert[e->verts[index / m]];
    index -= n;

    if (index < (n = (m = u->count_edge) * NEdge))
	return g->types_edge[e->edges[index / m]];
    index -= n;

    if (index < (n = (m = u->count_face) * NFace))
	return g->types_face[e->faces[index / m]];
    index -= n;

    return g->types_elem[e->index];
}

GTYPE
phgDofGetElementGeometryType(DOF *u, int index, int *eindex)
/* returns geometric type and index (in '*eindex') of the unknown of 'u'
 * with local 'index' on element 'e' */
{
    int m, n;

    if (index < (n = (m = u->count_vert) * NVert)) {
	*eindex = index / m;
	return VERTEX;
    }
    index -= n;

    if (index < (n = (m = u->count_edge) * NEdge)) {
	*eindex = index / m;
	return EDGE;
    }
    index -= n;

    if (index < (n = (m = u->count_face) * NFace)) {
	*eindex = index / m;
	return FACE;
    }
    index -= n;

    *eindex = index / u->count_elem;
    return ELEMENT;
}

FLOAT *
phgDofGetElementCoordinates(DOF *u, SIMPLEX *e, int index)
/* returns world coordinates of the unknown with local 'index' on element 'e' */
{
    static FLOAT c[Dim];
    GRID *g = u->g;
    FLOAT *c0, *c1, *c2, *c3, *L;
    int m, n;

    if (index < (n = (m = u->count_vert) * NVert))
	return g->verts[e->verts[index / m]];
    index -= n;

    if (u->type->points == NULL)
	phgError(1, "%s:%d, DOF points for DOF \"%s\" undefined.\n",
		 __FILE__, __LINE__, u->type->name);

    if (index < (n = (m = u->count_edge) * NEdge)) {
	L = u->type->points + 1 + ((index / u->dim) % u->type->np_edge) * 2;
	c0 = g->verts[e->verts[GetEdgeVertex(index /= m, 0)]];
	c1 = g->verts[e->verts[GetEdgeVertex(index, 1)]];
	c[0] = c0[0] * L[0] + c1[0] * L[1];
	c[1] = c0[1] * L[0] + c1[1] * L[1];
	c[2] = c0[2] * L[0] + c1[2] * L[1];
	return c;
    }
    index -= n;
    L = u->type->points + 1 + u->type->np_edge * 2;

    if (index < (n = (m = u->count_face) * NFace)) {
	L += ((index / u->dim) % u->type->np_face) * 3;
	c0 = g->verts[e->verts[GetFaceVertex(index /= m, 0)]];
	c1 = g->verts[e->verts[GetFaceVertex(index, 1)]];
	c2 = g->verts[e->verts[GetFaceVertex(index, 2)]];
	c[0] = c0[0] * L[0] + c1[0] * L[1] + c2[0] * L[2];
	c[1] = c0[1] * L[0] + c1[1] * L[1] + c2[1] * L[2];
	c[2] = c0[2] * L[0] + c1[2] * L[1] + c2[2] * L[2];
	return c;
    }
    index -= n;
    L += u->type->np_face * 3 + ((index / u->dim) % u->type->np_elem) * 4;

    c0 = g->verts[e->verts[0]];
    c1 = g->verts[e->verts[1]];
    c2 = g->verts[e->verts[2]];
    c3 = g->verts[e->verts[3]];
    c[0] = c0[0] * L[0] + c1[0] * L[1] + c2[0] * L[2] + c3[0] * L[3];
    c[1] = c0[1] * L[0] + c1[1] * L[1] + c2[1] * L[2] + c3[1] * L[3];
    c[2] = c0[2] * L[0] + c1[2] * L[1] + c2[2] * L[2] + c3[2] * L[3];
    return c;
}

/*----------------- functions for exchanging neighbour data ---------------*/

NEIGHBOUR_DATA *
phgGetNeighbourData(DOF *dof)
/* collect data from neighbours */
{
    NEIGHBOUR_DATA *nd;

    nd = phgAlloc(sizeof(*nd));
    nd->dof = dof;
    nd->size = dof->count_vert * NVert + dof->count_edge * NEdge +
	dof->count_face * NFace + dof->count_elem;
    nd->info = NULL;
    nd->data = NULL;

    return nd;
}

void
phgReleaseNeighbourData(NEIGHBOUR_DATA *nd)
{
    phgFree(nd->info);
    phgFree(nd->data);
    phgFree(nd);
}

FLOAT *
phgNeighbourData(NEIGHBOUR_DATA *nd, SIMPLEX *e, int neighbour, int type,
		 int index)
{
    DOF *dof = nd->dof;

    WARN_DOF_PERMUTATION(dof);
    assert(dof->type->dim == 1);

    e = (SIMPLEX *)e->neighbours[neighbour];
    switch (type) {
	case VERTEX:
	    return DofVertexData(dof, e->verts[index]);
	case EDGE:
	    return DofEdgeData(dof, e->edges[index]);
	case FACE:
	    return DofFaceData(dof, e->faces[index]);
	case ELEMENT:
	    return DofElementData(dof, e->index);
    }

    return NULL;
}
