/* Parallel Hierarchical Grid -- an adaptive finite element library.
 *
 * Copyright (C) 2005-2007 State Key Laboratory of Scientific and
 * Engineering Computing, Chinese Academy of Sciences. */

/* This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA */

/* Parallel I/O functions
 *
 * $Id: io.c,v 1.2 2006/03/09 06:45:14 zlb Exp $
 */

#include "phg.h"
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <limits.h>	/* PATH_MAX */

/* Use sequential I/O functions */

static FILE *fp = NULL;

BOOLEAN
phgIOOpen(GRID *g, const char *fn)
{
    if (g->rank > 0)
	return TRUE;
    return (fp = fopen(fn, "wb")) != NULL;
}

int
phgIOPrintf(GRID *g, char *fmt, ...)
/* serial fprintf analog to an MPI file: at first rank 0 writes the data,
   then new File View set to end of file */
{
    va_list ap;
    int ret;

    if (g->rank > 0)
	return 0; 

    va_start(ap, fmt);
    ret = vfprintf(fp, fmt, ap);
    va_end(ap);

    return ret;
}

void
phgIOWrite(GRID *g, void *buffer, size_t size, size_t count, size_t nglobal,
		BYTE flags[], INT map[])
/* Parallel data write.
 *
 * Each process has 'count' entries of size 'size' stored in 'buffer[]',
 * 'flags[i]' indicates whether entry i is owned by current process,
 * 'map[i]' is the global id of entry i 
 *
 * If flags == NULL then all entries are owned by current process
 * If map == NULL then map[i] == i + (sum of count of prev procs) is assumed.
 */
{

    assert(count == nglobal);
    fwrite(buffer, size, count, fp);
}

void
phgIOClose(void)
{
    if (fp != NULL) {
	fclose(fp);
	fp = NULL;
    }
}

const char *
phgIOAddExtension(const char *fn, const char *ext)
/* adds '.ext' to the given filename if the latter does not contain
 * an extension, returns the address of the new filename which points either
 * to 'fn' or a static buffer */
{
    static char new_name[PATH_MAX];
    const char *p;

    if (ext == NULL)
	return fn;
    if ((p = strrchr(fn, '/')) == NULL)
	p = fn;
    if ((p = strchr(p, '.')) != NULL)
	return fn;
    sprintf(new_name, "%s.%s", fn, ext);
    return new_name;
}
