/* Parallel Hierarchical Grid -- an adaptive finite element library.
 *
 * Copyright (C) 2005-2007 State Key Laboratory of Scientific and
 * Engineering Computing, Chinese Academy of Sciences. */

/* This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA */

/* $Id: map.c,v 1.72 2007/04/09 08:55:45 zlb Exp $ */

#define NEED_GetVariableArgs

/* The following macro controls whether DOFs in a vector are interleaved (1)
 * or concatenated (0).
 *
 * Warning: if INTERLEAVE_DOFS != 0 then the numbering of a multi-dof vector
 * is different when nprocs>1 then when nprocs==1. Also some applications
 * assume non interleaved ordering, so don't change its value unless for
 * testing purpose. */
#define INTERLEAVE_DOFS	0

#include "phg.h"

#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <stdarg.h>
#include <math.h>

/*------------ Functions managing maps ------------*/

static INT *tmp_map = NULL;	/* local to global geometric indices map */

static INT
get_global_index(MAP *map, INT index)
/* maps local index 'index' to global index */
{
    GRID *g;
    DOF *u;
    INT m, n;
    INT index0;
    int dof_no, ndof = map->ndof;

    if (map->nprocs <= 1)
	return index;

    /* determine the dof_no of the index */
    for (dof_no = 0; dof_no < ndof; dof_no++) {
	if (index < map->offset[dof_no + 1])
	    break;
    }
    if (dof_no >= ndof) {
	phgError(1, "get_global_index: invalid local index: %d\n", index);
    }

    index -= map->offset[dof_no];
    u = map->dofs[dof_no];
    g = u->g;

    index0 = 0;

    if (u->count_vert > 0) {
	if (index < (n = (m = u->count_vert) * g->nvert))
	    return tmp_map[index / m] + index % m + map->count_vert[dof_no];
	index -= n;
    }
    index0 = (map->count_vert[ndof] > 0) ? g->nvert : 0;

    if (u->count_edge > 0) {
	if (index < (n = (m = u->count_edge) * g->nedge))
	    return tmp_map[index0 + index / m] + index % m
					       + map->count_edge[dof_no];
	index -= n;
    }
    index0 += (map->count_edge[ndof] > 0) ? g->nedge : 0;

    if (u->count_face > 0) {
	if (index < (n = (m = u->count_face) * g->nface))
	    return tmp_map[index0 + index / m] + index % m
					       + map->count_face[dof_no];
	index -= n;
    }
    index0 += (map->count_face[ndof] > 0) ? g->nface : 0;

    assert(u->count_elem > 0);
    m = u->count_elem;
    return tmp_map[index0 + index / m] + index % m + map->count_elem[dof_no];
}

static void
build_L2Vmap(MAP *map)
{
    int ii;
    DOF *u = map->dofs[0];
    GRID *g = u->g;
    INT i, j, k, m, n, index0;
    INT count_vert, count_edge, count_face, count_elem;

    ii = (int)map->ndof + 1;
    map->offset = phgAlloc(ii * sizeof(*map->offset));
    map->offset[0] = 0;
    for (i = 0; i < map->ndof; i++) {
	u = map->dofs[i];
	map->offset[i + 1] = map->offset[i] + DofGetDataCount(u);
    }

    if (map->count_vert == NULL) {
	map->count_vert = phgAlloc(ii * sizeof(*map->count_vert));
	map->count_edge = phgAlloc(ii * sizeof(*map->count_edge));
	map->count_face = phgAlloc(ii * sizeof(*map->count_face));
	map->count_elem = phgAlloc(ii * sizeof(*map->count_elem));

	count_vert = count_edge = count_face = count_elem = 0;
	for (ii = 0; ii < map->ndof; ii++) {
	    u = map->dofs[ii];
	    map->count_vert[ii] = count_vert;
	    map->count_edge[ii] = count_edge;
	    map->count_face[ii] = count_face;
	    map->count_elem[ii] = count_elem;
	    count_vert += u->count_vert;
	    count_edge += u->count_edge;
	    count_face += u->count_face;
	    count_elem += u->count_elem;
	}
	map->count_vert[ii] = count_vert;
	map->count_edge[ii] = count_edge;
	map->count_face[ii] = count_face;
	map->count_elem[ii] = count_elem;
    }
    else {
	count_vert = map->count_vert[ii];
	count_edge = map->count_edge[ii];
	count_face = map->count_face[ii];
	count_elem = map->count_elem[ii];
    }

    if (g->nprocs <= 1) {
	map->lif = 1.0;
	map->partition = phgAlloc(2 * sizeof(*map->partition));
	map->partition[0] = 0;
	map->partition[1] = map->nlocal;
	goto skip;
    }

  skip:
    map->L2Vmap = phgAlloc(map->localsize * sizeof(*map->L2Vmap));
    map->ordering = NULL;
    if (map->localsize == map->nlocal)
	map->O2Gmap = NULL;
    else
	map->O2Gmap = phgAlloc((map->localsize - map->nlocal)
				  * sizeof(*map->O2Gmap));
    n = map->nlocal;
    i = 0;
    index0 = map->partition[map->rank];
    for (j = 0; j < map->ndof; j++) {
	u = map->dofs[j];
	m = DofGetDataCount(u);
	for (k = 0; k < m; k++, i++) {
	    INT l = get_global_index(map, i);
	    if (l >= index0 && l < index0 + map->nlocal) {
		map->L2Vmap[i] = l - index0;
	    }
	    else {
		map->O2Gmap[n - map->nlocal] = l;
		map->L2Vmap[i] = n++;
	    }
	}
    }
    assert(n == map->localsize);

    if (tmp_map != NULL)
	phgFree(tmp_map);

    return;
}

INT
phgMapL2G(MAP *map, INT index)
{
    DOF *u = map->dofs[0];
    INT i;

    assert(index >= 0 && index < map->localsize);

    if (u->g->serial_no != map->serial_no)
	phgError(1, "phgMapL2G: unexpected grid change, abort.\n");

    if (u->g->nprocs == 1)
	return index;

    i = (map->L2Vmap == NULL ? index : map->L2Vmap[index]);
    assert(i >= 0 && i < map->localsize);

    if (i < map->nlocal)
	return map->partition[map->rank] + i;
    else
	return map->O2Gmap[i - map->nlocal];
}

static INT
get_local_size(MAP *map)
{
    GRID *g = (map->dofs == NULL ? NULL : map->dofs[0]->g);
    DOF *u;
    INT i, n, size = 0;
    int j;

    for (j = 0; j < map->ndof; j++) {
	u = map->dofs[j];

	if ((n = u->count_vert) > 0)
	    for (i = 0; i < g->nvert; i++) {
		if (g->types_vert[i] & OWNER)
		    size += n;
	    }

	if ((n = u->count_edge) > 0)
	    for (i = 0; i < g->nedge; i++) {
		if (g->types_edge[i] & OWNER)
		    size += n;
	    }

	if ((n = u->count_face) > 0)
	    for (i = 0; i < g->nface; i++) {
		if (g->types_face[i] & OWNER)
		    size += n;
	    }

	size += u->count_elem * g->nleaf;
    }

    return size;
}

static void
get_bdry_map(MAP *map)
/* initialize map->bdry_* members */
{
    GRID *g;
    DOF *dof;
    int ii, n;
    INT *counts, count;
    INT i, j, k, index, index0, gindex0;

    BTYPE b, *tmp_types[4];
    int tmp_count[4];
    INT tmp_size[4];
    BTYPE DB_mask;

    if (map->dofs == NULL)
	return;

    g = map->dofs[0]->g;

    for (i = 0; i < map->ndof; i++)
        if (map->dofs[i]->DB_mask != 0)
	    break;
    if (i >= map->ndof)
	return;

    gindex0 = map->partition[g->rank];

    tmp_types[0] = g->types_vert;
    tmp_types[1] = g->types_edge;
    tmp_types[2] = g->types_face;
    tmp_types[3] = g->types_elem;

    tmp_size[0] = g->nvert;
    tmp_size[1] = g->nedge;
    tmp_size[2] = g->nface;
    tmp_size[3] = g->nelem;

    /* count number of boundary DOF owned by the present process.
     * Note: by putting the loop on map->dofs as the inner most one,
     * then entries in map->bdry[] are increasing */
    count = 0;
    phgFree(map->bdry);
    map->bdry = NULL;
    map->bdry_nlocal = 0;
    for (i = 0; i < map->ndof; i++) {
	dof = map->dofs[i];
	DB_mask = dof->DB_mask;
	if (DB_mask == 0)
	    continue;
	tmp_count[0] = dof->count_vert;
	tmp_count[1] = dof->count_edge;
	tmp_count[2] = dof->count_face;
	tmp_count[3] = dof->count_elem;
	index0 = 0;
	for (ii = 0; ii < 4; ii++) {
	    if ((n = tmp_count[ii]) == 0)
		continue;
	    for (j = 0; j < tmp_size[ii]; j++) {
		if (!((b = tmp_types[ii][j]) & DB_mask))
		    continue;
		if (count + n > map->bdry_nlocal) {
		    map->bdry_nlocal = (count + n + 1024);
		    map->bdry = phgRealloc(map->bdry, map->bdry_nlocal *
							sizeof(*map->bdry));
		}
		for (index = index0 + n * j;
		     index < index0 + n * (j + 1); index++) {
		    k = map->L2Vmap[phgMapD2L(map, i, index)];
		    assert(k < map->localsize);
#if 0
		    /* only mark local bdry entries */
		    if (k >= map->nlocal)
			continue;
#endif
		    map->bdry[count++] = k;
		}
	    }
	    index0 += tmp_size[ii] * n;
	}
    }

    if (count > 0) {
	/* sort entries in bdry[] */
	qsort(map->bdry, count, sizeof(*map->bdry), phgCompINT);
    }

    map->bdry_localsize = count;
    for (i = 0; i < count && map->bdry[i] < map->nlocal; i++);
    map->bdry_nlocal = count = i;

    /* gather counts from all processes */
    counts = phgAlloc(g->nprocs * sizeof(*counts));
    k = counts[0] = count;
    map->bdry_partition[0] = 0;
    map->bdry_partition[1] = count;
    map->bdry_nglobal = k;

    phgInfo(2, "nlocal: %d, boundary DOF: %d (%d total)\n",
	    map->nlocal, count, k);

    phgFree(counts);

#if 0
    k = map->partition[map->rank];
    phgInfo(-1, "n0 = %d, nlocal = %d, localsize = %d, nglobal = %d\n",
		    k, map->nlocal, map->localsize, map->nglobal);
    phgInfo(-1, "bdry_nlocal = %d\n", map->bdry_nlocal);
    for (i = 0; i < map->bdry_nlocal; i++) {
	assert(map->bdry[i] < map->nlocal);
	assert(i == 0 || map->bdry[i] > map->bdry[i - 1]);
	phgInfo(-1, "bdry[%d] = %d\n", i, k + map->bdry[i]);
    }
    phgInfo(-1, "bdry_localsize = %d\n", map->bdry_localsize);
    for (; i < map->bdry_localsize; i++) {
	k = map->bdry[i] - map->nlocal;
	assert(i == map->bdry_nlocal || map->bdry[i] > map->bdry[i - 1]);
	assert(k >= 0);
	phgInfo(-1, "bdry[%d] = %d (%d)\n", i, map->O2Gmap[k], k);
    }
    phgInfo(-1, "\n");
    if (map->localsize == map->nlocal)
	return;
    j = map->O2Gmap[map->ordering[0]];
    for (i = 1; i < map->localsize - map->nlocal; i++) {
	k = map->O2Gmap[map->ordering[i]];
	assert(k > j);
	j = k;
    }
#endif
}

MAP *
phgMapCreateSimpleMap(GRID *g, INT n, INT N)
{
    MAP *map = phgCalloc(1, sizeof(*map));
    int i;

    assert(n != PHG_DECIDE || N != PHG_DECIDE);

    map->nprocs = g->nprocs;
    map->rank = g->rank;
    map->serial_no = g->serial_no;		/* save serial number */

    if (n == PHG_DECIDE) {
	n = N / map->nprocs;
	if (map->rank < N % map->nprocs)
	    n++;
    }
    assert(N == PHG_DECIDE || n <= N);
    map->nlocal = n;
    map->partition = phgAlloc((map->nprocs + 1) * sizeof(*map->partition));
    map->partition[0] = 0;
    map->partition[1] = n;
    for (i = 0; i < map->nprocs; i++)
	map->partition[i+ 1] += map->partition[i];
    map->nglobal = map->partition[i];
    if (N == PHG_DECIDE)
	N = map->nglobal;

    if (map->nglobal > N) {
	/* special case: identical serial map on all processes */
	assert(map->nglobal == map->nprocs * N);
	map->nglobal = N;
	map->nprocs = 1;
	map->rank = 0;
	phgFree(map->partition);
	map->partition = phgAlloc(2 * sizeof(*map->partition));
	map->partition[0] = 0;
	map->partition[1] = N;
    }

    assert(map->nglobal == N);
    map->localsize = map->nlocal;

    map->bdry_nlocal = map->bdry_nglobal = map->bdry_localsize = 0;
    map->bdry = NULL;
    map->bdry_partition =
	phgCalloc(map->nprocs + 1, sizeof(*map->bdry_partition));

    return map;
}

MAP *
phgMapCreateN(int ndof, DOF **dofs)
{
    MAP *map = phgCalloc(1, sizeof(*map));
    GRID *g;
    DOF *u;
    int j;

    assert(dofs != NULL);

    map->ndof = ndof;
    map->dofs = phgAlloc(ndof * sizeof(*map->dofs));
    memcpy(map->dofs, dofs, ndof * sizeof(*map->dofs));
    u = dofs[0];
    g = u->g;
    map->nprocs = g->nprocs;
    map->rank = g->rank;
    map->serial_no = g->serial_no;		/* save serial number */

    map->nlocal = get_local_size(map);
    map->nglobal = map->localsize = 0;
    for (j = 0; j < map->ndof; j++) {
	u = map->dofs[j];
	map->nglobal += DofGetDataCountGlobal(u);
	map->localsize += DofGetDataCount(u);
    }

    build_L2Vmap(map);

    map->bdry_nlocal = map->bdry_nglobal = map->bdry_localsize = 0;
    map->bdry = NULL;
    map->bdry_partition =
	phgCalloc(g->nprocs + 1, sizeof(*map->bdry_partition));
    get_bdry_map(map);

    return map;
}

MAP *
phgMapCreate(DOF *u, ...)
{
    int ndof;
    DOF **dofs;
    MAP *map;

    GetVariableArgs(ndof, dofs, u, DOF *);
    map = phgMapCreateN(ndof, dofs);
    phgFree(dofs);

    return map;
}

void
phgMapDestroy(MAP **map_ptr)
{
    MAP *map = *map_ptr;

    if (map == NULL)
	return;

    *map_ptr = NULL;

    phgMapDestroy(&map->map_nb);

    assert(map->ref_count >= 0);

    if (map->ref_count > 0) {
	map->ref_count--;
	return;
    }

    phgFree(map->L2Vmap);
    phgFree(map->O2Gmap);
    phgFree(map->ordering);
    phgFree(map->partition);
    phgFree(map->offset);
    phgFree(map->bdry);
    phgFree(map->bdry_partition);

    phgFree(map->count_vert);
    phgFree(map->count_edge);
    phgFree(map->count_face);
    phgFree(map->count_elem);

    phgFree(map->dofs);

    phgFree(map);

    return;
}

MAP *
phgMapRemoveBoundaryEntries(MAP *map)
/* derives a new map from 'map' with all boundary entries removed */
{
    INT n, N;

    if (map->bdry_nglobal == 0) {
	map->ref_count++;
	return map;
    }

    if (map->map_nb != NULL) {
	map->map_nb->ref_count++;
	return map->map_nb;
    }
 
    n = map->nlocal - map->bdry_nlocal;
    N = map->nglobal - map->bdry_nglobal;
    assert(map->dofs != NULL);
    map->map_nb = phgMapCreateSimpleMap(map->dofs[0]->g, n, N);
    map->map_nb->ref_count++;	/* note: map_nb is referenced by map */

    return map->map_nb;
}

/*--------- functions for vector scattering/gathering ---------*/

/*--------- functions for transferring data between DOF list and VEC --------*/

int
phgMapLocalDataToDof(MAP *map, int ndof, DOF **dofs, FLOAT *data)
/* This function copies the solution back to DOFs.
 * 'vec' is the pointer to the local vector data (size = map->nlocal)
 */
{
#if 1	/*==================================================================*/
  /* The old circular algorithm */
  {
    DOF *u;
    INT i, j, k = 0, l, m, nmax;
    INT nlocal = map->nlocal;
    /* struct for storing remote data */
    struct {
	FLOAT value;
	INT gindex;
    } *rdata = NULL;
    INT *offp_counts = NULL;

    nmax = 0;
    if (offp_counts != NULL) {
	for (i = 0; i < map->nprocs; i++) {
	    if (nmax < offp_counts[i])
		nmax = offp_counts[i];
	}
    }
    if (nmax == 0) {
	k = 0;
	goto skip_comm;
    }

  skip_comm:
    /* copy data to DOFs */
    for (l = 0; l < ndof; l++) {
	u = dofs[l];
	m = DofGetDataCount(u);
	for (i = 0; i < m; i++) {
	    j = map->L2Vmap[phgMapD2L(map, l, i)];
	    if (j < nlocal) {
		u->data[i] = data[j];
	    }
	    else {
		u->data[i] = rdata[k + j - nlocal].value;
	    }
	}
    }

    phgFree(offp_counts);
    phgFree(rdata);

    return 0;
  }
#endif	/*==================================================================*/
	
#if 0
  {
    int id;
    VEC *vec = phgMapCreateVec(map, 0);
    DOF ***pdofs = phgAlloc(ndof * sizeof(*pdofs));

    for (id = 0; id < ndof; id++)
	pdofs[id] = &dofs[id];
    vec->nvec = 1;
    vec->data = data;
    phgMapVecToDofArraysN(map, vec, FALSE, ndof, pdofs);
    vec->nvec = 0;
    vec->data = NULL;
    phgVecDestroy(&vec);
    phgFree(pdofs);
  }
#else	/* 0|1 */
  {
    int id;
    INT i, k, m;
    FLOAT *offp_data;
    DOF *u;

    assert(map->dofs != NULL);
    assert(ndof == map->ndof);

    offp_data = phgAlloc((map->localsize - map->nlocal) * sizeof(*data));

    assert(map->localsize == 0 || map->L2Vmap != NULL);
    for (id = 0; id < ndof; id++, dofs++) {
	u = dofs[id];
	m = DofGetDataCount(u);
	for (i = 0; i < m; i++) {
	    k = map->L2Vmap[phgMapD2L(map, id, i)];
	    if (k < map->nlocal)
		u->data[i] = data[k];
	    else
		u->data[i] = offp_data[k - map->nlocal];
	}
    }

    phgFree(offp_data);
  }
#endif	/* 0|1 */

    return 0;
}

int
phgMapDofToLocalData(MAP *map, int ndof, DOF **dofs, FLOAT *data)
/* This function copies the data in the DOFs to local vector 
 * array (for e.g., setting up an initial solution of a solver).
 * 'data' is the pointer to the local data (size = map->nlocal)
 */
{
    DOF *u;
    int k;
    INT i, j, m;
    INT nlocal = map->nlocal;

    assert(map->localsize == 0 || map->L2Vmap != NULL);
    for (k = 0; k < ndof; k++) {
	u = dofs[k];
	m = DofGetDataCount(u);
	for (i = 0; i < m; i++) {
	    j = map->L2Vmap[phgMapD2L(map, k, i)];
	    if (j < nlocal) {
		data[j] = u->data[i];
	    }
	}
    }

    return 0;
}

VEC *
phgMapDofArraysToVecN(MAP *map, int nvec, BOOLEAN remove_bdry,
			 VEC **vecptr, int ndof, DOF ***dofs)
/* copies data from an array of DOFs to a vector */
{
    VEC *vec = (vecptr == NULL ? NULL : *vecptr);
    int id, iv;
    DOF *u, *u0;
    INT i, j, k, m, *table;
    FLOAT *data;

    assert(map->dofs != NULL);
    assert(ndof == map->ndof);
    MagicCheck(VEC, vec);

    if (vec != NULL) {
	phgVecDestroy(&vec);
	if (vecptr != NULL)
	    *vecptr = NULL;
    }

    if (nvec <= 0)
	return vec;

    /* expand bdry[] for faster checking */
    if (!remove_bdry) {
	/* no boundary entries */
	vec = phgMapCreateVec(map, nvec);
	table = phgAlloc(map->localsize * sizeof(*table));
	for (i = 0; i < map->localsize; i++)
	    table[i] = i;
    }
    else {
	MAP *map0 = phgMapRemoveBoundaryEntries(map);
	vec = phgMapCreateVec(map0, nvec);
	phgMapDestroy(&map0);
	table = phgCalloc(map->localsize, sizeof(*table));
	for (i = 0; i < map->bdry_localsize; i++)
	    table[map->bdry[i]] = 1;
	for (i = 0, k = 0; i < map->localsize; i++) {
	    if (table[i] == 1)
		table[i] = -1;
	    else
		table[i] = k++;
	}
    }

    if (vecptr != NULL)
	*vecptr = vec;

    assert(map->localsize == 0 || map->L2Vmap != NULL);
    for (id = 0; id < ndof; id++, dofs++) {
	u0 = (*dofs)[0];
	m = DofGetDataCount(u0);
	for (i = 0; i < m; i++) {
	    k = table[j = map->L2Vmap[phgMapD2L(map, id, i)]];
	    if (k == -1 || k >= vec->map->nlocal)
		continue;
	    data = vec->data + k;
	    for (iv = 0; iv < nvec; iv++) {
		u = (*dofs)[iv];
		assert(u->type == u0->type && u->dim == u0->dim);
		*data = u->data[i];
		data += vec->map->nlocal;
	    }
	}
    }
    phgFree(table);

    return vec;
}

VEC *
phgMapDofArraysToVec(MAP *map, int nvec, BOOLEAN remove_bdry,
			VEC **vecptr, DOF **u, ...)
/* copies data from an array of DOFs to a vector */
{
    int ndof;
    VEC *vec;
    DOF ***dofs;

    GetVariableArgs(ndof, dofs, u, DOF **);
    vec = phgMapDofArraysToVecN(map, nvec, remove_bdry, vecptr, ndof, dofs);
    phgFree(dofs);
    return vec;
}

void
phgMapVecToDofArraysN(MAP *map, VEC *vec, BOOLEAN remove_bdry,
			 int ndof, DOF ***dofs)
/* copies data from a vector to an array of DOFs */
{
    int id, iv, nvec;
    INT i, j, k, m, stride;
    FLOAT *data, *offp_data = NULL;
    static FLOAT zero = 0.0;
    DOF *u0, *u;
    INT *table;

    assert(map->dofs != NULL);
    assert(ndof == map->ndof);
    assert(vec != NULL);
    MagicCheck(VEC, vec);

    if (vec->nvec <= 0)
	return;

    nvec = vec->nvec;

    /* expand bdry[] for faster checking */
    if (!remove_bdry || vec->map->nglobal == map->nglobal) {
	/* no boundary entries */
	assert(vec->map->nglobal   == map->nglobal &&
	       vec->map->localsize == map->localsize);
	table = phgAlloc(map->localsize * sizeof(*table));
	for (i = 0; i < map->localsize; i++)
	    table[i] = i;
    }
    else {
	assert(vec->map->nlocal    == map->nlocal    - map->bdry_nlocal &&
	       vec->map->nglobal   == map->nglobal   - map->bdry_nglobal);
	table = phgCalloc(map->localsize, sizeof(*table));
	for (i = 0; i < map->bdry_localsize; i++)
	    table[map->bdry[i]] = 1;
	for (i = 0, k = 0; i < map->localsize; i++) {
	    if (table[i] == 1)
		table[i] = -1;
	    else
		table[i] = k++;
	}
    }

    assert(map->localsize == 0 || map->L2Vmap != NULL);
    for (id = 0; id < ndof; id++, dofs++) {
	u0 = (*dofs)[0];
	m = DofGetDataCount(u0);
	for (i = 0; i < m; i++) {
	    k = table[j = map->L2Vmap[phgMapD2L(map, id, i)]];
	    if (k == -1) {
		data = &zero;
		stride = 0;
	    }
	    else if (k < vec->map->nlocal) {
		data = vec->data + k;
		stride = vec->map->nlocal;
	    }
	    else {
		assert(j >= map->nlocal);
		data = offp_data + j - map->nlocal;
		stride = map->localsize - map->nlocal;
	    }
	    for (iv = 0; iv < nvec; iv++) {
		u = (*dofs)[iv];
		assert(u->type == u0->type && u->dim == u0->dim);
		u->data[i] = *data;
		data += stride;
	    }
	}
    }

    phgFree(offp_data);
    phgFree(table);
}

void
phgMapVecToDofArrays(MAP *map, VEC *vec, BOOLEAN remove_bdry, DOF **u, ...)
/* copies data from a vector to an array of DOFs */
{
    int ndof;
    DOF ***dofs;

    GetVariableArgs(ndof, dofs, u, DOF **);
    phgMapVecToDofArraysN(map, vec, remove_bdry, ndof, dofs);
    phgFree(dofs);
}

VEC *
phgMapCreateVec(MAP *map, int nvec)
{
    VEC *v = phgCalloc(1, sizeof(*v));

    map->ref_count++;

    v->magic = MAGIC_VEC;
    v->map = map;
    v->assembled = FALSE;
    v->nvec = nvec;
    v->data = phgCalloc(nvec * map->nlocal, sizeof(*(v->data)));
    v->offp_data = phgCalloc(nvec * (map->localsize - map->nlocal),
						sizeof(*(v->data)));

    return v;
}

MAT *
phgMapCreateMat(MAP *rmap, MAP *cmap)
{
    MAT *Mp;

    FunctionEntry;

    Mp = phgCalloc(1, sizeof(*Mp));
    Mp->magic = MAGIC_MAT;
    Mp->rmap = rmap;
    Mp->cmap = cmap;
    Mp->type = PHG_UNPACKED;

    /* check for special matrix */
    if (Mp->rmap->nprocs == 1 && Mp->cmap->nprocs > 1) {
	phgPrintf("\n*\n* Warning: Matrices with serial rows and distributed "
		  "columns are inefficient.\n");
	phgPrintf("* Warning: It is better to use distributed rows and serial "
		  "columns instead.\n*\n");
    }

    Mp->localsize = -1;

    Mp->rows = phgCalloc(rmap->localsize, sizeof(*Mp->rows));
    Mp->nnz_d = Mp->nnz_o = 0;

    Mp->diag = NULL;
    Mp->packed_data = NULL;
    Mp->packed_cols = NULL;

    Mp->mv_func = NULL;
    Mp->mv_data = NULL;

    rmap->ref_count++;
    cmap->ref_count++;

    Return Mp;
}

MAT *
phgMapCreateMatrixFreeMat(MAP *rmap, MAP *cmap, MV_FUNC mv_func,
			  void *mv_data, ...)
{
    extern void _phg_matrix_free_setup(MAT *, MV_FUNC, int, void **);
    MAT *mat = phgMapCreateMat(rmap, cmap);
    int ndata;
    void **data;

    if (mat == NULL)
	return NULL;

    GetVariableArgs(ndata, data, mv_data, void *);
    _phg_matrix_free_setup(mat, mv_func, ndata, data);
    phgFree(data);

    return mat;
}
