/* Parallel Hierarchical Grid -- an adaptive finite element library.
 *
 * Copyright (C) 2005-2007 State Key Laboratory of Scientific and
 * Engineering Computing, Chinese Academy of Sciences. */

/* This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA */

/* $Id: utils.c,v 1.152 2007/05/22 08:58:36 zlb Exp $ */

#define _GNU_SOURCE	/* for fcloseall() */

#include "phg.h"

#include <stdlib.h>
#include <unistd.h>	/* usleep */
#include <limits.h>	/* PATH_MAX */
#include <stdarg.h>
#include <string.h>
#include <math.h>

#include <sys/time.h>
#ifdef USETIMES
# include <sys/times.h>
# include <unistd.h>
#else
# include <sys/resource.h>
#endif
#define TRUNK_SIZE 8

#if HAVE_FEENABLEEXCEPT
# include <fenv.h>
#endif	/* HAVE_FEENABLEEXCEPT */
#include <signal.h>

INT phgVerbosity = 0;
const char *phgCurrentFunctionName_ = "PHG";
INT _phg_i = 0;			/* scratch INT variable used in some macros */

static GRID *g_traverse;	/* used by phgTraverseElements, other functions
				   shouldn't use it */
static GRID *g_;		/* for traversal callback functions */
BOOLEAN phgInitialized = FALSE;
BOOLEAN phgMasterSlave = FALSE;

char *phgLogFilename = NULL;
char *phgOutFilename = NULL;
static FILE *log_file = NULL;
static FILE *out_file = NULL;

#define LogFile	(log_file == NULL ? stderr : log_file)
#define OutFile	(out_file == NULL ? stdout : out_file)

static void
init_log_file(void)
{
    char s[PATH_MAX];

    if (phgLogFilename == NULL || log_file != NULL)
	return;

    if (strcmp(phgLogFilename, "stdout") == 0) {
	log_file = stdout;
	return;
    }
    else if (strcmp(phgLogFilename, "stderr") == 0) {
	log_file = stderr;
	return;
    }

    if (phgNProcs > 1)
	sprintf(s, "%s.%04d", phgLogFilename, phgRank);
    else
	strcpy(s, phgLogFilename);
    
    log_file = fopen(s, "w+t");
    if (log_file == NULL) {
	phgFree(phgLogFilename);
	phgLogFilename = NULL;
    }
}

static void
close_log_file(void)
{
    if (log_file != NULL && log_file != stdout && log_file != stderr) {
	fclose(log_file);
	if (phgRank > 0)
	    return;
	if (phgNProcs > 1)
	    phgPrintf("Log files \"%s.%04d\"--\"%s.%04d\" created.\n",
			phgLogFilename, 0, phgLogFilename, phgNProcs - 1);
	else
	    phgPrintf("Log file \"%s\" created.\n", phgLogFilename);
    }
    log_file = NULL;

    fflush(OutFile);
    if (out_file != NULL && out_file != stdout && out_file != stderr) {
	fclose(out_file);
    }
    out_file = NULL;
}

int
phgCompINT(const void *p0, const void *p1)
{
    INT i = *((INT *)p0) - *((INT *)p1);

    return (i < 0) ? -1 : ((i == 0) ? 0 : 1);
}

INT
phgBinarySearch(INT n, INT table[], INT key)
/* binary search, table[] is a sorted table of size n.
 *   1. if (n == 0 || key < table[0]) return -1;
 *   2. if (key > table[n - 1]) return n;
 *   3.	return i, s.t. (key >= table[i] && (i == n-1 || key < table[i + 1]); */
{
    INT i = 0, j = n - 1, k;

    if (n <= 0 || key < table[0])
	return -1;
    if (key >= table[j])
	return n - 1;

    while (j - i > 1) {
	k = (i + j) / 2;
	assert(k != i && k != j);
	if (table[k] <= key)
	    i = k;
	else
	    j = k;
    }

    return i;
}

#if HAVE_FEENABLEEXCEPT
static void
fpe_handler(int signum, siginfo_t *siginfo, void *dummy)
{
    switch (siginfo->si_code) {
	case FPE_FLTDIV:
	    phgError(1, "floating point exception: divide by zero.\n");
	    break;
	case FPE_FLTOVF:
	    phgError(1, "floating point exception: overflow.\n");
	    break;
	case FPE_FLTUND:
	    phgError(1, "floating point exception: underflow.\n");
	    break;
	case FPE_FLTRES:
	    phgError(1, "floating point exception: inexact result.\n");
	    break;
	case FPE_FLTINV:
	    phgError(1, "floating point exception: invalid operation.\n");
	    break;
	default:
	    phgError(1, "program received signal SIGFPE, abort.\n");
    }
}
#endif	/* HAVE_FEENABLEEXCEPT */

void
phgTrapSignals_(BOOLEAN init)
/* Note: should be called once with 'init == TRUE' once before phgInit,
 * and once with 'init == FALSE' to set FPE trap after initialization
 * of all external packages (like PETSc) */
{
#if HAVE_FEENABLEEXCEPT
    int flags = 0;
    static const char *fpetrap_keywords[] = {"none", "fatal", "all", NULL};
    static int fpetrap = 1;
#endif	/* HAVE_FEENABLEEXCEPT */
    static BOOLEAN reset_signals = TRUE;
    struct sigaction sa;

    if (init) {
	/* Note: the function is called once with 'init == TRUE' for
	 * registering options, then once with 'init == FALSE' after
	 * initialization of external solvers, to make sure the sigactions
	 * are called after external packages */
	phgOptionsRegisterNoArg("reset_signals", "Reset signal handlers",
		&reset_signals);
#if HAVE_FEENABLEEXCEPT
	phgOptionsRegisterKeyword("fpetrap",
		"Trap floating point exceptions",
		fpetrap_keywords, &fpetrap);
#endif	/* HAVE_FEENABLEEXCEPT */
	return;
    }

    if (reset_signals) {
	/* reset some signal handlers to default */
	sa.sa_handler = SIG_DFL;
	sigemptyset(&sa.sa_mask);
	sa.sa_flags = 0;
	if (sigaction(SIGPIPE, &sa, NULL))
	    perror("sigaction");
	if (sigaction(SIGINT, &sa, NULL))
	    perror("sigaction");
	if (sigaction(SIGQUIT, &sa, NULL))
	    perror("sigaction");
	if (sigaction(SIGILL, &sa, NULL))
	    perror("sigaction");
	if (sigaction(SIGSEGV, &sa, NULL))
	    perror("sigaction");
	if (sigaction(SIGTERM, &sa, NULL))
	    perror("sigaction");
    }

#if HAVE_FEENABLEEXCEPT
    switch (fpetrap) {
	case 2:	/* all */
	    flags = FE_UNDERFLOW /*| FE_INEXACT*/;
	    /* fall through the next case */
	case 1:	/* fatal */
	    flags |= FE_DIVBYZERO | FE_INVALID | FE_OVERFLOW;
	    /* Clear current exceptions queue.
	     * Note: the initialization code of the Anasazi package from
	     * Trilinos-7.0.5 generates 'invalid operation' traps on i386 */
	    feclearexcept(flags);
	    if (feenableexcept(flags) == -1) {
		perror("");
		exit(1);
	    }
	    break;
	case 0:	/* none */
	    return;
    }

    /* install SIGFPE handler */
    sa.sa_sigaction = fpe_handler;
    sigemptyset(&sa.sa_mask);
    sa.sa_flags = SA_SIGINFO;
    if (sigaction(SIGFPE, &sa, NULL))
	perror("sigaction");
#endif	/* HAVE_FEENABLEEXCEPT */
}

BOOLEAN _phg_pause = FALSE;

void
phgInit(int *argc, char ***argv)
{
    if (phgInitialized) return;

    /* test validity of basic datatypes */
    assert(sizeof(INT) >= 4);
    assert(sizeof(UINT) >= 4);
    assert(sizeof(SHORT) >= 2);
    assert(sizeof(USHORT) >= 2);
    assert(sizeof(BYTE) == 1);

    /* Register marking options */
    phgMarkElements_(0, NULL, 0., NULL, 0., 0, 0., 0, NULL, 0., 0, 0.);
    /* Register perf options */
    phgPerfInit();
    /* Register solver options */
    phgSolverRegisterOptions();

    phgPrintf("Parallel Hierarchical Grid (version " PHG_VERSION ").\n");
    phgOptionsParseCmdline(argc, argv);

    /* Initialize modules after parsing all options.
     * Note: also add the calls below in phgInitMPI() in phg-mpi.c */
    phgPerfInit();			/* init phgPerf */
    phgSolverInitialize(argc, argv);	/* init phgSolver */

    phgOptionsHelp();

    if (phgVerbosity >= 2) {
	phgPrintf("sizeof(SIMPLEX)    = %d\n", sizeof(SIMPLEX));
	phgPrintf("sizeof(GRID)       = %d\n", sizeof(GRID));
    }
    fflush(OutFile);

    phgTrapSignals_(FALSE);
    phgTimerInit(FALSE);

    phgInitialized = TRUE;
    if (_phg_pause)
	phgPause(0);
    if (phgMasterSlave) atexit(phgFinalize);
}

int
phgSetVerbosity(int verbosity)
{
    int old = phgVerbosity;
    if (verbosity >= 0) phgVerbosity = verbosity;
    return old;
}

void
phgFinalize(void)
{
    if (!phgInitialized)
	return;

#if USE_PAPI
    phgPerfFinalize();
#endif
    phgTimerReset();

    phgSolverFinalize();
    phgQuadReset();

    close_log_file();

    phgOptionsReset();

    phgInitialized = FALSE;

    phgFreeAtExit(NULL);

    if (phgRank > 0)
	exit(0);
}

void
phgPause(int seconds)
{
    char s[128];

    if (seconds <= 0) {
	int i;
	pid_t pid = getpid(), *pids = phgAlloc(phgNProcs * sizeof(*pids));
	pids[0] = pid;
	if (phgRank == 0) {
            fprintf(stderr, "Program paused.\n");
	    for (i = 0; i < phgNProcs; i++) {
		fprintf(stderr, "  proc %d, pid = %d\n", i, (int)pids[i]);
	    }
	}
	phgFree(pids);
    }
    if (phgRank == 0) {
	if (seconds <= 0) {
            fflush(stderr);
            fprintf(stderr, "<<< Press enter to continue >>>\n");
            fflush(stderr);
            fgets(s, sizeof(s), stdin);
            fflush(stderr);
	}
	else {
	    sleep(seconds);
	}
    }
}

static void
init_elements(SIMPLEX *s, int count)
{
    int i, j;

    for (i = 0; i < count; i++) {
	for (j = 0; j < NEdge; j++) {
	    s[i].edges[j] = -1;
#if ALLOW_CURVED_BOUNDARY
	    s[i].bound_func[j] = -1;
#endif
	}
	for (j = 0; j < NFace; j++) {
	    s[i].bound_type[j] = UNDEFINED;
#if (Dim == 3)
	    s[i].faces[j] = -1;
#endif
	}
	s[i].index = -1;
    }
}

SIMPLEX *
phgNewElements(int count)
{
    SIMPLEX *s = phgCalloc(count, sizeof(SIMPLEX));

    init_elements(s, count);
    return s;
}

SIMPLEX *
phgReallocElements(SIMPLEX *s, int oldcount, int newcount)
{
    if (newcount == oldcount)
	return s;
    s = phgRealloc(s, newcount * sizeof(SIMPLEX));
    if (newcount <= oldcount)
	return s;
    memset(s + oldcount, 0, (newcount - oldcount) * sizeof(*s));
    init_elements(s + oldcount, newcount - oldcount);

    return s;
}

COORD *
phgNewVertices(int count)
{
    COORD *v;

    v = phgAlloc(count * sizeof(COORD));
    return v;
}

static int flags_tmp;
static size_t g_list_count = 0;
static GRID **g_list = NULL;	/* list of valid grids */

void
phgNewGrid_(GRID *g)
{
    ParallelFunction(phgNewGrid_, TRUE);

    /* for Master/Slave mode in which only one grid is allowed */
    g_sys = g;
    if (g != NULL) return;

    g_sys = g = phgCalloc(1, sizeof(GRID));
    g->lif = 1.0;
    g->nelem = g->nleaf = g->nroot = g->nvert = g->nedge = g->ntree = 0;
#if Dim == 3
    g->nface = 0;
    g->nface_global = 0;
#endif
    g->verts = NULL;
    g->roots = NULL;
    g->leaves = NULL;
    g->dof   = NULL;
    g->vert_flag = /*flags_tmp & VERT_FLAG*/TRUE; /* always required */
    g->edge_flag = flags_tmp & EDGE_FLAG;
    g->face_flag = flags_tmp & FACE_FLAG;
    g->elem_flag = flags_tmp & ELEM_FLAG;
    g->geom_flag = flags_tmp & GEOM_FLAG;
    g->periodic = FALSE;
#if ALLOW_CURVED_BOUNDARY
    g->bdry_funcs = NULL;
#endif	/* ALLOW_CURVED_BOUNDARY */
    g->types_vert = NULL;
    g->types_edge = NULL;
    g->types_face = NULL;
    g->types_elem = NULL;
    g->perlist = NULL;
    g->nelem_global = g->nvert_global = g->nedge_global = 0;
    g->rank = 0;
    g->nprocs = 1;
    g->serial_no = 0;

    /* add g to g_list (for phgGridIsValid) */
    g_list = phgRealloc(g_list, (g_list_count + 1) * sizeof(*g_list));
    g_list[g_list_count++] = g;

    return;
}

GRID *
phgNewGrid(int flags)
{
    flags_tmp = flags;
    phgNewGrid_(NULL);

    return g_sys;
}

BOOLEAN
phgGridIsValid(const GRID *g)
/* checks whether the given grid is valid */
{
    int i;

    for (i = 0; i < g_list_count; i++)
	if (g_list[i] == g) return TRUE;

    return FALSE;
}

/* Array mapping edge no to vertices, used by GetEdgeVertex macro */
BYTE GetEdgeVertex_[NEdge][2] = {{0,1}, {0,2}, {0,3}, {1,2}, {1,3}, {2,3}};
/* Array mapping face no to vertices, used by GetFaceVertex macro */
BYTE GetFaceVertex_[NFace][3] = {{1,2,3}, {0,2,3}, {0,1,3}, {0,1,2}};
/* array mapping pairs of vertices to edges */
BYTE GetEdgeNo_[NVert][NVert] = {{-1, 0, 1, 2},
				 { 0,-1, 3, 4},
				 { 1, 3,-1, 5},
				 { 2, 4, 5,-1}};

static INT index_ = 0;

static BOOLEAN
leaves_callback CB_ARGS(e)
{
    int v0, v1, v2, v3;
    INT V0, V1, V2, V3;

    if (IsLeaf(e)) {
	g_->leaves[index_++] = e;
    }
    else {
	/* point e->parent to parent */
	if (e->children[0] != NULL)
	    e->children[0]->parent = e;
	if (e->children[1] != NULL)
	    e->children[1]->parent = e;
    }
    if (e->generation == 0)
	e->parent = NULL;

    if (e->ordering != 0)
	return TRUE;

    /* compute ordering of vertices with increasing global indices */
    V0 = GlobalVertex(g_, e->verts[0]);
    V1 = GlobalVertex(g_, e->verts[1]);
    V2 = GlobalVertex(g_, e->verts[2]);
    V3 = GlobalVertex(g_, e->verts[3]);
    v0 = v1 = v2 = v3 = 0;
    (V0 > V1) ? v0++ : v1++;
    (V0 > V2) ? v0++ : v2++;
    (V0 > V3) ? v0++ : v3++;
    (V1 > V2) ? v1++ : v2++;
    (V1 > V3) ? v1++ : v3++;
    (V2 > V3) ? v2++ : v3++;
    e->ordering = v0 | (v1 << 2) | (v2 << 4) | (v3 << 6);

    return TRUE;
}

static BOOLEAN
types_callback CB_ARGS(e)
{
    int i, type;

    if (g_->elem_flag)
	g_->types_elem[e->index] = OWNER;

    for (i = 0; i < NFace; i++) {
	type = e->bound_type[i];
#if 0
	/* use DIRICHLET for UNDEFINED boundary type */
	if (type == UNDEFINED) type = DIRICHLET;
#endif
	type |= OWNER;

	if (g_->vert_flag) {
	    switch (i) {
		case 0:
		    g_->types_vert[e->verts[1]] |= type;
		    g_->types_vert[e->verts[2]] |= type;
		    g_->types_vert[e->verts[3]] |= type;
		    break;
		case 1:
		    g_->types_vert[e->verts[0]] |= type;
		    g_->types_vert[e->verts[2]] |= type;
		    g_->types_vert[e->verts[3]] |= type;
		    break;
		case 2:
		    g_->types_vert[e->verts[0]] |= type;
		    g_->types_vert[e->verts[1]] |= type;
		    g_->types_vert[e->verts[3]] |= type;
		    break;
		case 3:
		    g_->types_vert[e->verts[0]] |= type;
		    g_->types_vert[e->verts[1]] |= type;
		    g_->types_vert[e->verts[2]] |= type;
		    break;
	    }
	}

	if (g_->edge_flag) {
	    switch (i) {
		case 0:
		    g_->types_edge[e->edges[3]] |= type;
		    g_->types_edge[e->edges[4]] |= type;
		    g_->types_edge[e->edges[5]] |= type;
		    break;
		case 1:
		    g_->types_edge[e->edges[1]] |= type;
		    g_->types_edge[e->edges[2]] |= type;
		    g_->types_edge[e->edges[5]] |= type;
		    break;
		case 2:
		    g_->types_edge[e->edges[0]] |= type;
		    g_->types_edge[e->edges[2]] |= type;
		    g_->types_edge[e->edges[4]] |= type;
		    break;
		case 3:
		    g_->types_edge[e->edges[0]] |= type;
		    g_->types_edge[e->edges[1]] |= type;
		    g_->types_edge[e->edges[3]] |= type;
		    break;
	    }
	}

	if (g_->face_flag) {
	    g_->types_face[e->faces[i]] |= type;
	}

	if (g_->elem_flag) {
	    g_->types_elem[e->index] |= (type & REMOTE);
	}
    }

    return TRUE;
}

void
phgUpdateBoundaryTypes(GRID *g)
/* sets boundary types (g->types_vert and g->types_edge).
 * This function also updates g->leaves */
{
    SIMPLEX *e;
    INT i;

    phgFree(g->leaves);
    g->leaves = NULL;

    if (g->nprocs <= phgRank)
	return;

    g_ = g;
    g->leaves = phgAlloc(g->nleaf * sizeof(*g->leaves));
    index_ = 0;
    /* construct g->leaves, update e->parent and e->odering */
    phgTraverseAllElements(g, leaves_callback);

    if (!g->vert_flag && !g->edge_flag && !g->face_flag && !g->elem_flag)
	return;

    if (g->vert_flag) {
	phgFree(g->types_vert);
	g->types_vert = phgAlloc(g->nvert * sizeof(*g->types_vert));
	memset(g->types_vert, UNREFERENCED, g->nvert * sizeof(*g->types_vert));
    }

    if (g->edge_flag) {
	phgFree(g->types_edge);
	g->types_edge = phgAlloc(g->nedge * sizeof(*(g->types_edge)));
	memset(g->types_edge, UNREFERENCED, g->nedge);
    }

    if (g->face_flag) {
	phgFree(g->types_face);
	g->types_face = phgAlloc(g->nface * sizeof(*(g->types_face)));
	memset(g->types_face, UNREFERENCED, g->nface);
    }

    if (g->elem_flag) {
	phgFree(g->types_elem);
	g->types_elem = phgAlloc(g->nelem * sizeof(*(g->types_elem)));
	memset(g->types_elem, UNREFERENCED, g->nelem);
    }

    ForAllElements(g, e)
	types_callback(e);

    if (g->vert_flag)
	for (i = 0; i < g->nvert; i++)
	    g->types_vert[i] |= OWNER;
    if (g->edge_flag)
	for (i = 0; i < g->nedge; i++)
	    g->types_edge[i] |= OWNER;
    if (g->face_flag)
	for (i = 0; i < g->nface; i++)
	    g->types_face[i] |= OWNER;

    if (g->elem_flag)
	for (i = 0; i < g->nelem; i++)
	    g->types_elem[i] |= OWNER;
}

void *
phgAlloc(size_t size)
{
    void *ptr = (size != 0) ? malloc(size) : NULL;
    
    if (ptr == NULL && size != 0)
	phgError(1, "failed to malloc %u bytes memory.\n", size);

    return ptr;
}

void *
phgCalloc(size_t nmemb, size_t size)
{
    void *ptr = (nmemb != 0 && size != 0) ? calloc(nmemb, size) : NULL;

    if (ptr == NULL && nmemb != 0 && size != 0)
	phgError(1, "failed to calloc %u bytes memory.\n", size * nmemb);

    return ptr;
}

void *
phgRealloc(void *ptr, size_t size)
{
    void *p = realloc(ptr, size);
    
    if (p == NULL && size != 0)
	phgError(1, "failed to reallocate %u bytes memory.\n", size);

    return p;
}

void
phgFree(void *ptr)
{
    if (ptr != NULL) free(ptr);
}

int
phgMemcmp(const void *v0, const void *v1, size_t n)
/* replaces buggy memcmp() on many systems */
{
    const BYTE *s0 = v0, *s1 = v1;
    int i = 0;

    for (; n--; )
	if ((i = *(s0++) - *(s1++)) != 0)
	    break;

    return i;
}

void
phgFreeAtExit(void **ptr)
{
    static void ***ptr_table = NULL;
    static size_t table_n = 0, allocated = 0;

    if (ptr == NULL) {	/* free all registered pointers */
	int i;
	for (i = table_n - 1; i >= 0; i--)
	    phgFree(*(ptr_table[i]));
	phgFree(ptr_table);
	ptr_table = NULL;
	table_n = allocated = 0;

	return;
    }

    if (table_n >= allocated)
	ptr_table = (void***)phgRealloc(ptr_table,
				sizeof(*ptr_table) * (allocated += 128));
    ptr_table[table_n++] = ptr; 

    return;
}

double
phgGetTime(double tarray[])
/* tarray[0]=User time, tarray[1]=Sys time, tarray[2]=Elapsed time,
   all expressed in seconds.

   DAWN1000 nf77
	double F77name(dclock)();
	tarray[0]=tarray[1]=tarray[2]=F77name(dclock)(); */
{
    struct timeval tv;
  
    if (tarray != NULL) {
#ifdef USETIMES  /*==================================================*/
	struct tms ts;
	static double ClockTick = 0.0;

	if (ClockTick == 0.0)
	    ClockTick = (double)sysconf(_SC_CLK_TCK);
	if ( times(&ts) == -1 ) {
	    tarray[0]=tarray[1] = 0.0;
	} else {
	    tarray[0] = (double)ts.tms_utime / ClockTick;
	    tarray[1] = 0.0;
	}
#else		 /*==================================================*/
	/* For Solaris */
	/*#include </usr/ucbinclude/sys/rusage.h> */
	struct rusage RU;
	getrusage(RUSAGE_SELF, &RU);
	tarray[0] = RU.ru_utime.tv_sec + (double)RU.ru_utime.tv_usec * 1e-6;
	tarray[1] = RU.ru_stime.tv_sec + (double)RU.ru_stime.tv_usec * 1e-6;
#endif		 /*==================================================*/
    }

    gettimeofday(&tv, (struct timezone *)0);
    if (tarray != NULL)
	return tarray[2] = tv.tv_sec + (double)tv.tv_usec * 1e-6;
    else
	return tv.tv_sec + (double)tv.tv_usec * 1e-6;
}

void
phgWarning(const char *fmt, ...)
{
    va_list ap;
    char s[1024];
    const char *n;

    init_log_file();

    n = phgCurrentFunctionName_;
    if (strlen(n) >= 3 && !phgMemcmp(n, "phg", 3)) n += 3;

    if (phgNProcs > 1)
	sprintf(s, "*%d %s: *** WARNING: %s", phgRank, n, fmt);
    else
	sprintf(s, "* %s: *** WARNING: %s", n, fmt);
    va_start(ap, fmt);
    vfprintf(LogFile, s, ap);
    va_end(ap);
    fflush(LogFile);
}

void
phgAbort(int code)
{
    close_log_file();

    /* leave time for processes to clean up */
    if (phgNProcs > 1)
	usleep(500000);

#ifdef HAVE_FCLOSEALL
    fcloseall();
#endif

    exit(code);
}

void 
phgError(int code, const char *fmt, ...)
{
    va_list ap;
    char s[4096], pn[1];
    const char *n;

    init_log_file();

    n = phgCurrentFunctionName_;
    if (strlen(n) >= 3 && !phgMemcmp(n, "phg", 3)) n += 3;

    pn[0] = '\0';
    if (phgNProcs > 1)
	sprintf(s, "\n*%d%s %s: *** ERROR: %s\n", phgRank, pn, n, fmt);
    else
	sprintf(s, "\n*%s %s: *** ERROR: %s\n", pn, n, fmt);
    va_start(ap, fmt);
    vfprintf(LogFile, s, ap);
    va_end(ap);
    fflush(LogFile);

    if (code == 0) return;
    phgAbort(code);
}

void
phgInfo(int verbose_level, const char *fmt, ...)
/* writes formatted message to log_file when log_file != NULL,
 * if verbose_level < 0 then writes to stdout if log_file == NULL */
{
    va_list ap;
    char s[1024];
    const char *n;

    init_log_file();

    if ((verbose_level >= 0 && log_file == NULL) ||
	verbose_level > phgVerbosity) return;

    n = phgCurrentFunctionName_;
    if (strlen(n) >= 3 && !phgMemcmp(n, "phg", 3)) n += 3;

    if (phgNProcs > 1)
	sprintf(s, "*%d %s: %s", phgRank, n, fmt);
    else
	sprintf(s, "* %s: %s", n, fmt);
    va_start(ap, fmt);
    vfprintf(LogFile, s, ap);
    va_end(ap);
    fflush(LogFile);
}

#undef Log

int
phgPrintf(const char *fmt, ...)
{
    va_list ap;
    int ret = 0;

    if (phgRank == 0) {
	if (out_file == NULL && phgOutFilename != NULL) {
	    if (strcmp(phgOutFilename, "stdout") == 0) {
		out_file = stdout;
	    }
	    else if (strcmp(phgOutFilename, "stderr") == 0) {
		out_file = stderr;
	    }
	    else {
		out_file = fopen(phgOutFilename, "w+t");
		if (out_file == NULL) {
		    phgWarning("cannot open output file \"%s\".",
							phgOutFilename);
		    phgFree(phgOutFilename);
		    phgOutFilename = NULL;
		}
	    }
	}
	va_start(ap, fmt);
	ret = vfprintf(OutFile, fmt, ap);
	va_end(ap);
	fflush(OutFile);
    }

    return ret;
}

FILE *
phgFOpen(GRID *g, const char *path, const char *mode)
{
    Unused(g);
    return fopen(path, mode);
}

int
phgFPrintf(GRID *g, FILE *fp, char *fmt, ...)
{
    va_list ap;
    int ret;

    Unused(g);
    va_start(ap, fmt);
    ret = vfprintf(fp, fmt, ap);
    va_end(ap);
    fflush(fp);

    return ret;
}

int
phgFClose(GRID *g, FILE *fp)
{
    int ret;

    Unused(g);
    ret = fclose(fp);

    return ret;
}

BYTE
phgOppositeVertex(SIMPLEX *e, INT v, SIMPLEX *e_op)
/* return the vertex of e_op opposite to the vertex v (local index) of e.
   This function does not rely on the neighbours links */
{
    INT i, u0 = 0, u1 = 0, u2 = 0;

    switch (v) {
	case 0: u0 = e->verts[1]; u1 = e->verts[2]; u2 = e->verts[3]; break;
	case 1: u0 = e->verts[0]; u1 = e->verts[2]; u2 = e->verts[3]; break;
	case 2: u0 = e->verts[0]; u1 = e->verts[1]; u2 = e->verts[3]; break;
	case 3: u0 = e->verts[0]; u1 = e->verts[1]; u2 = e->verts[2]; break;
    }

    if ((i = e_op->verts[0]) != u0 && i != u1 && i != u2)
	return 0;
    if ((i = e_op->verts[1]) != u0 && i != u1 && i != u2)
	return 1;
    if ((i = e_op->verts[2]) != u0 && i != u1 && i != u2)
	return 2;
    if ((i = e_op->verts[3]) != u0 && i != u1 && i != u2)
	return 3;
    return 4;
}

void
phgFreeElement(SIMPLEX **eptr)
{
    SIMPLEX *e;

    if (eptr == NULL || (e = *eptr) == NULL)
	return;
#if 0
    if (e->userdata != NULL)
	phgFree(e->userdata);
#endif
    phgFree(e);
    *eptr = NULL;
}

static void
free_grid_traverse(SIMPLEX *s)
{
    if (s->children[0] != NULL)
	free_grid_traverse(s->children[0]);
    if (s->children[1] != NULL)
	free_grid_traverse(s->children[1]);
    phgFreeElement(&s);
}

void
phgFreeGrid_(GRID *g)
{
    INT i, n;

    ParallelFunction(phgFreeGrid_, g != NULL && g->nprocs > 1);

    if (g == NULL)
	return;

    for (i = 0; i < g->nroot; i++) {
	SIMPLEX *s = g->roots + i;
	if (s->children[0] != NULL)
	    free_grid_traverse(s->children[0]);
	if (s->children[1] != NULL)
	    free_grid_traverse(s->children[1]);
#if 0
	if (s->userdata != NULL)
	    phgFree(s->userdata);
#endif
    }
    phgFree(g->roots);
    phgFree(g->verts);
    phgFree(g->leaves);
#if ALLOW_CURVED_BOUNDARY
    if (g->bdry_funcs != NULL) {
	EXPR **p = g->bdry_funcs;
	while (*p != NULL)
	    phgFree3DFunction(*(p++));
	phgFree(g->bdry_funcs);
    }
#endif
    phgFree(g->types_vert);
    phgFree(g->types_edge);
    phgFree(g->types_face);
    phgFree(g->types_elem);
    phgFree(g->perlist);

    /* avoid warning on unfreed DOF */
    if (g->geom != NULL)
	phgDofFree(&g->geom);

    /* Free DOFs attached to the grid */
    if (g->dof != NULL) {
	DOF **p;
	for (p = g->dof; *p != NULL; p++);
	i = (INT)(p - g->dof);
	while (i > 0) {
	    DOF *d = g->dof[--i];
#if DEBUG
	    /* For helping finding out memory leaks due to unfreed DOFs */
	    phgPrintf("*** WARNING: unfreed DOF \"%s\" (created at %s:%d)\n",
			d->name, d->srcfile, d->srcline);
#endif
	    d->g = NULL; /* make phgGridIsValid return FALSE in phgDofFree */
	    phgDofFree(&d);
	}
	phgFree(g->dof);
    }
    phgFree(g);

    /* remove g from g_list */
    for (i = 0; i < g_list_count; i++)
	if (g_list[i] == g) break;

    if (i >= g_list_count)
	phgError(1, "phgFreeGrid: invalid grid!\n");

    n = g_list_count - i - 1;
    if (n > 0) {
	memmove(g_list + i, g_list + i + 1, n * sizeof(*g_list));
    }
    if (--g_list_count > 0) {
	g_list = phgRealloc(g_list, g_list_count * sizeof(*g_list));
    }
    else {
	phgFree(g_list);
	g_list = NULL;
    }

    /* for Master/Slave mode in which only one grid is allowed */
    g_sys = NULL;

    return;
}

void
phgFreeGrid(GRID **gptr)
{
    GRID *g = (gptr == NULL ? NULL : *gptr);

    if (g == NULL)
	return;

    if (gptr != NULL)
	*gptr = NULL;

    phgFreeGrid_(g);
}

/* flag controlling whether to call callback function for non-leaf elements */
static BOOLEAN traverse_all_flag = FALSE;

/* Depth of the element in the tree.
   This variable is made global and may be used by the callback functions */
INT phgTraverseDepth = 0, phgTraverseIndex = 0;

static BOOLEAN
traverse_element(SIMPLEX *e, BOOLEAN(*cb) CB_ARGS(e))
{
    BOOLEAN is_leaf = TRUE;

    if (e->children[0] != NULL) {
        is_leaf = FALSE;
	phgTraverseDepth++;
	if (!traverse_element(e->children[0], cb)) return FALSE;
	phgTraverseDepth--;
    }
    if (e->children[1] != NULL) {
        is_leaf = FALSE;
	phgTraverseDepth++;
	if (!traverse_element(e->children[1], cb)) return FALSE;
	phgTraverseDepth--;
    }
    if (!is_leaf && !traverse_all_flag) {
	return TRUE;
    }
    else {
	BOOLEAN ret = cb CB_ARGS0(g_traverse, e);
	Unused(g_traverse);
	phgTraverseIndex++;
	return ret;
    }
}

void
phgTraverseElements(GRID *g, BOOLEAN(*cb) CB_ARGS(e))
/* function traversing (traverse_all_flag ? leaf : all) elements.
 * Stop traversing if callback function returns FALSE */
{
    INT i;

    if (g->nroot == 0) return;

    g_traverse = g;
    phgTraverseIndex = 0;
    for (i = 0; i < g->nroot; i++) {
	phgTraverseDepth = 0;
	if (!traverse_element(g->roots + i, cb)) break;
    }
}

void
phgTraverseAllElements(GRID *g, BOOLEAN(*cb) CB_ARGS(e))
/* function traversing all (including non leaf) elements */
{
    traverse_all_flag = TRUE;
    phgTraverseElements(g, cb);
    traverse_all_flag = FALSE;
}

static BOOLEAN conformity; /* used to check conformity during traversal */

static struct {
    SIMPLEX **elements;
    SHORT   nelem, nmax;
} *vertices, *pv, *pv0[Dim];

static BOOLEAN
find_neighbours_callback0 CB_ARGS(s)
{
    INT k;
    int j;

    memset(s->neighbours, 0, sizeof(s->neighbours));

    for (j = 0; j < NVert; j++) {
	k = s->verts[j];
	pv = vertices + k;
	if (pv->nelem >= pv->nmax)
	    pv->elements = phgRealloc(pv->elements,
				(pv->nmax+=32) * sizeof(s));
	pv->elements[pv->nelem++] = s;
    }

    if (s->index == -1)
	s->index = phgTraverseIndex;

    return TRUE;
}

#define EXTENSIVE_TEST 0

#if DEBUG && EXTENSIVE_TEST
static BOOLEAN
singular_solve(int n, int m, double a[n][m], double b[n])
/* solves the singular system AX == B (A is a singular NxM matrix),
 * returns FALSE if no solution */
{
    int i, j, k;
    double d, e;

    assert(m <= n);

    /* normalize the coefficients */
    for (i = 0; i < n; i++) {
	d = fabs(a[i][0]);
	for (j = 1; j < m; j++) {
	    if (d < (e = fabs(a[i][j])))
		d = e;
	}
	if (d == 0.)
	    continue;
	d = 1. / d;
	for (j = 1; j < m; j++)
	    a[i][j] *= d;
	b[i] *= d;
    }

    /* Gauss elimination */
    for (i = 0; i < m; i++) {
	k = i;
	d = fabs(a[i][i]);
	for (j = i + 1; j < n; j++) {
	    if (d < (e = fabs(a[j][i]))) {
		d = e;
		k = j;
	    }
	}
	if (d < 1e-12)
	    return FALSE;
	if (k != i) {
	    /* exchange row i with row k */
	    d = b[i];
	    b[i] = b[k];
	    b[k] = d;
	    for (j = 0; j < m; j++) {
		d = a[i][j];
		a[i][j] = a[k][j];
		a[k][j] = d;
	    }
	}
	/* elimination */
	d = 1. / a[i][i];
	for (j = 0; j < m; j++)
	    a[i][j] *= d;
	b[i] *= d;
	for (j = i + 1; j < n; j++) {
	    for (k = i + 1; k < m; k++)
		a[j][k] -= a[j][i] * a[i][k];
	    b[j] -= a[j][i] * b[i];
	}
    }

    /* check consistency of the system */
    for (; i < n; i++) {
	if (fabs(b[i]) > 1e-12)
	    return FALSE;
    }

    /* back substitute */
    for (i = m - 2; i >= 0; i--)
	for (j = i + 1; j < m; j++)
	    b[i] -= a[i][j] * b[j];

    return TRUE;
}

static BOOLEAN
check_intersect(COORD *v0, COORD *v1, COORD *u0, COORD *u1)
/* returns TRUE if the segments v0-v1 and u0-u1 intersect */
{
    int i;
    double a[Dim][2], b[Dim];

    for (i = 0; i < Dim; i++) {
	b[i] = (*u0)[i] - (*v0)[i];
	a[i][0] = (*v1)[i] - (*v0)[i];
	a[i][1] = -((*u1)[i] - (*u0)[i]);
    }

    if (!singular_solve(Dim, 2, a, b))
	return FALSE;

    if (b[0] < -1e-12 || b[0] > 1 + 1e-12 || b[1] < -1e-12 || b[1] > 1 + 1e-12)
	return FALSE;

    return TRUE;
}

static void
check2(SIMPLEX *s, int face, int fa, int fb, int fc, SIMPLEX *s1)
/* check the case in which 's1' has two vertices on the give face of 's'
 * (vertices 'fa' and 'fb' are shared by 's' and 's1').
 *
 * This check handles a special case in which two edges of two faces
 * sharing two common vertices intersect. This typically happens when
 * a hexahedron is adjacent to a tetrahedron and is tetrahedralized
 * inconsistently */
{
    int j;
    INT v0, v1, v2, v3;
    COORD *coor;

    if (IS_BDRY(s->bound_type[face]))
	return;

    v0 = s->verts[GetFaceVertex(face, fa)],
    v1 = s->verts[GetFaceVertex(face, fb)],
    v2 = s->verts[GetFaceVertex(face, fc)];
    coor = g_->verts;

    /* make sure edges of s1 does not cross the face */
    for (j = 0; j < NVert; j++) {
	if ((v3 = s1->verts[j]) == v0 || v3 == v1)
	    continue;
	assert(v3 != v2);
	/* (v3-v0) should not intersect with (v2-v1) and
	   (v3-v1) should not intersect with (v2-v0) */
	if (check_intersect(coor + v3, coor + v0, coor + v2, coor + v1) ||
	    check_intersect(coor + v3, coor + v1, coor + v2, coor + v0))
	    phgError(1, "nonconforming mesh, abort.\n");
    }
}

typedef struct {
    SIMPLEX *e;
    BYTE v;
} ELIST;

static int
comp_elist(const void *p0, const void *p1)
{
    const ELIST *el0 = p0, *el1 = p1;
    return (el0->e > el1->e ? 1 : (el0->e == el1->e ? 0 : -1));
}

#endif	/* DEBUG && EXTENSIVE_TEST */

static BOOLEAN
find_neighbours_callback1 CB_ARGS(s)
{
#if 0
    /* new code based on sorting the list of neighbours of local vertices.
     * FIXME: the code is much slower and breaks mesh coarsening algorithm */
    SIMPLEX *s1;
    struct {
	INT	eno;
	BYTE	v;
    } *elist = NULL;
    int i, j, n, m;
    INT eno;

    /* count number of elements in the lists of the 3 vertices */
    n = 0;
    for (j = 0; j < NVert; j++) {
	s->neighbours[j] = NULL;
	n += vertices[s->verts[j]].nelem;
    }
    n -= NVert;
    if (n <= 0)
	goto cont;
    elist = phgAlloc(n * sizeof(*elist));
    /* collect element indices */
    n = 0;
    for (j = 0; j < NVert; j++) {
	for (i = 0; i < vertices[s->verts[j]].nelem; i++) {
	    if ((s1 = vertices[s->verts[j]].elements[i]) == s)
		continue;
	    elist[n].eno = s1 - g_->roots;
	    elist[n++].v = j;
	}
    }
    qsort(elist, n, sizeof(*elist), phgCompINT);

    eno = -1;	/* current element */
    m = 0;	/* count of duplicate elements */
    for (j = 0; j <= n; j++) {
	if (j < n && eno == elist[j].eno) {
	    m++;
	    continue;
	}
	assert(m <= 3);
	if (m == 3) {
	    BYTE v0 = elist[j - 3].v, v1 = elist[j - 2].v, v2 = elist[j - 1].v;
	    /* find the face number */
	    if (v0 != 0 && v1 != 0 && v2 != 0)
		i = 0;
	    else if (v0 != 1 && v1 != 1 && v2 != 1)
		i = 1;
	    else if (v0 != 2 && v1 != 2 && v2 != 2)
		i = 2;
	    else
		i = 3;
	    s->bound_type[i] &= ~(REMOTE | UNDEFINED);
	    s->bound_type[i] |= INTERIOR;
	    s->neighbours[i] = g_->roots + eno;
	}
	else if (m == 2) {
	    /* TODO: further conformity check */
	}
	if (j >= n)
	    break;
	eno = elist[j].eno;
	m = 1;
    }

cont:
    for (i = 0; i < NVert; i++) {
	if (s->neighbours[i] != NULL)
	    continue;
	if (s->bound_type[i] & INTERIOR) {
	    if (g_->nprocs == 1) {
		conformity = FALSE;
	    }
	    else {
		s->bound_type[i] |= REMOTE;
		s->neighbours[i] = NULL;
	    }
	}
    }

    phgFree(elist);
#else	/* #if 0|1 */
    SIMPLEX *s1 = NULL;
    int face, j, l;
    INT k;

    for (face = 0; face < NFace; face++) {
	k = 0;
	for (j = 0; j < NVert; j++) {
	    if (j == face)
		continue;
	    pv0[k++] = vertices + s->verts[j];
	}
	/* element appearing in all the Dim lists is the neighbour */
	for (k = 0; k < pv0[0]->nelem; k++) {
	    s1 = pv0[0]->elements[k];
	    if (s1 == s)
		continue;
	    for (l = 0; l < pv0[1]->nelem; l++)
		if (s1 == pv0[1]->elements[l])
		    break;
	    if (l >= pv0[1]->nelem)
		continue;
#if Dim == 3
	    for (l = 0; l < pv0[2]->nelem; l++)
		if (s1 == pv0[2]->elements[l])
		    break;
	    if (l >= pv0[2]->nelem)
		continue;
#endif	/* Dim == 3 */
	    break;
	}
	if (k < pv0[0]->nelem) {
	    s->bound_type[face] &= ~(REMOTE | UNDEFINED);
	    s->bound_type[face] |= INTERIOR;
	    s->neighbours[face] = s1;
	}
	else {
	    if (s->bound_type[face] & INTERIOR) {
		if (g_->nprocs == 1) {
		    conformity = FALSE;
		}
		else {
		    s->bound_type[face] |= REMOTE;
		    s->neighbours[face] = NULL;
		}
	    }
	    /* already cleared by callback0() */
	    /* s->neighbours[face] = NULL;*/
	}
#if DEBUG && EXTENSIVE_TEST	/* Slow! don't enable unless for debugging */
	/* Check intersected edges */
	{
	    static BOOLEAN warned = FALSE;
	    ELIST *elist = NULL;
	    int i, n, m;

	    if (!warned) {
		warned = TRUE;
		phgWarning("slow code! don't enable except for debugging.\n");
	    }

	    /* count number of elements in the lists of the 3 vertices */
	    n = pv0[0]->nelem + pv0[1]->nelem + pv0[2]->nelem - 3;
	    if (n <= 0)
		continue;
	    elist = phgAlloc(n * sizeof(*elist));
	    /* collect element indices */
	    n = 0;
	    for (j = 0; j < 3; j++) {
		for (i = 0; i < pv0[j]->nelem; i++) {
		    if ((s1 = pv0[j]->elements[i]) == s)
			continue;
		    elist[n].e = s1;
		    elist[n++].v = j;
		}
	    }
	    qsort(elist, n, sizeof(*elist), comp_elist);
	
	    s1 = NULL;	/* current element */
	    m = 0;	/* count of duplicate elements */
	    for (j = 0; j <= n; j++) {
		if (j < n && s1 == elist[j].e) {
		    m++;
		    continue;
		}
		assert(m <= 3);
		if (m == 2) {
		    int v0 = elist[j - 1].v, v1 = elist[j - 2].v, v2; 
		    if (((v2 = 0) != v0 && v2 != v1) ||
			((v2 = 1) != v0 && v2 != v1) ||
			((v2 = 2) != v0 && v2 != v1))
			check2(s, face, v0, v1, v2, s1);
		}
		if (j >= n)
		    break;
		s1 = elist[j].e;
		m = 1;
	    }
	}
#endif /* DEBUG && EXTENSIVE_TEST */
    }
#endif	/* #if 0|1 */

    return TRUE;
}
	
static BOOLEAN
check_conformity_callback CB_ARGS(e)
{
    int i, j, k;
    SIMPLEX *e1;

    for (i = 0; i < NVert; i++) {
	/* check boundary types */
#if 0
	if (e->bound_type[i] == 0) {
	    phgInfo(-1, "undefined boundray type for face %d\n", i);
	    phgDumpElement(g_, e);
	    return conformity = FALSE;
	}
#endif
	/* check indices of vertices */
	if (e->verts[i] < 0 || e->verts[i] >= g_->nvert) {
	    phgInfo(-1, "invalid vertex index %d\n", e->verts[i]);
	    phgDumpElement(g_, e);
	    return conformity = FALSE;
	}
	/* check links to neighbours */
	if (!HasLocalNeighbour(e, i))
	    continue;
	e1 = e->neighbours[i];
	if (e1 == NULL) {
	    phgInfo(-1, "interior face (%d) without neighbour:\n", i);
	    phgDumpElement(g_, e);
	    if (i > 1)
		phgDumpPatch(g_, e, 0, 1);
	    else
		phgDumpPatch(g_, e, 2, 3);
	    return conformity = FALSE;
	}
	if (!IsLeaf(e1)) {
	    phgInfo(-1, "non-root neighbour:\n");
	    phgDumpElement(g_, e);
	    phgDumpElement(g_, e1);
	    if (e1->children[0] != NULL)
		phgDumpElement(g_, e1->children[0]);
	    if (e1->children[1] != NULL)
		phgDumpElement(g_, e1->children[1]);
	    return conformity = FALSE;
	}
	/* check that the vertices of the common face (edge in 2D) match */
	for (j = 0; j < NVert; j++) {
	    INT v;
	    if (j == i) continue;
	    v = e->verts[j];
	    for (k = 0; k < NFace; k++) {
		if (v == e1->verts[k]) break;
	    }
	    if (k >= NFace) {
		phgInfo(-1, "non conforming face:\n");
		phgDumpElement(g_, e);
		phgDumpElement(g_, e1);
		return conformity = FALSE;
	    }
	}
#if Dim == 3
	if (HasLocalNeighbour(e, i) && g_->face_flag) {
	    j = phgOppositeVertex(e, i, e1 = e->neighbours[i]);
	    if (e->faces[i] != e1->faces[j]) {
		phgInfo(-1, "non conforming face indices: %d <-> %d\n", i, j);
		phgDumpElement(g_, e);
		phgDumpElement(g_, e1);
		return conformity = FALSE;
	    }
	}
#endif
    }

    /* TODO: check edge indices */

    return TRUE;
}

static BOOLEAN
check_conformity_nonleaf_callback CB_ARGS(e)
{
    int fmap[NVert], emap[NEdge];
    SIMPLEX *e0, *e1;

    /* TODO: check vertices */

    if ((e0 = e->children[0]) != NULL) {
	if (g_->elem_flag && e->index != e0->index &&
	    GlobalVertex(g_, e->verts[0]) < GlobalVertex(g_, e->verts[1])) {
	    phgDumpElement(g_, e);
	    phgDumpElement(g_, e0);
	    phgInfo(-1, "index of child 0 != index of parent\n");
	    return conformity = FALSE;
	}

	if (g_->edge_flag || g_->face_flag)
	    phgMapP2C(e, fmap, emap, 0);

	if (g_->edge_flag &&
	    (e->edges[1] != e0->edges[emap[1]] ||
	     e->edges[2] != e0->edges[emap[2]] ||
	     e->edges[5] != e0->edges[emap[5]] ||
	     (GlobalVertex(g_, e->verts[0]) < GlobalVertex(g_, e->verts[1]) &&
	      e->edges[0] != e0->edges[emap[0]]))) {
	    phgDumpElement(g_, e);
	    phgDumpElement(g_, e0);
	    phgInfo(-1, "edge indices of parent and child 0 mismatch\n");
	    return conformity = FALSE;
	}

	if (g_->face_flag &&
	    (e->faces[1] != e0->faces[3] ||
	     (GlobalVertex(g_, e->verts[0]) < GlobalVertex(g_, e->verts[1]) &&
	      (e->faces[2] != e0->faces[fmap[2]] ||
	       e->faces[3] != e0->faces[fmap[3]])))) {
	    phgDumpElement(g_, e);
	    phgDumpElement(g_, e0);
	    phgInfo(-1, "face indices of parent and child 0 mismatch\n");
	    return conformity = FALSE;
	}
    }

    if ((e1 = e->children[1]) != NULL) {
	if (g_->elem_flag && e->index != e1->index &&
	    GlobalVertex(g_, e->verts[0]) > GlobalVertex(g_, e->verts[1])) {
	    phgDumpElement(g_, e);
	    phgDumpElement(g_, e0);
	    phgInfo(-1, "index of child 0 != index of parent\n");
	    return conformity = FALSE;
	}

	if (g_->edge_flag || g_->face_flag)
	    phgMapP2C(e, fmap, emap, 1);

	if (g_->edge_flag &&
	    (e->edges[3] != e1->edges[emap[3]] ||
	     e->edges[4] != e1->edges[emap[4]] ||
	     e->edges[5] != e1->edges[emap[5]] ||
	     (GlobalVertex(g_, e->verts[0]) > GlobalVertex(g_, e->verts[1]) &&
	      e->edges[0] != e1->edges[emap[0]]))) {
	    phgDumpElement(g_, e);
	    phgDumpElement(g_, e1);
	    phgInfo(-1, "edge indices of parent and child 1 mismatch\n");
	    return conformity = FALSE;
	}

	if (g_->face_flag &&
	    (e->faces[0] != e1->faces[3] ||
	     (GlobalVertex(g_, e->verts[0]) > GlobalVertex(g_, e->verts[1]) &&
	      (e->faces[2] != e1->faces[fmap[2]] ||
	       e->faces[3] != e1->faces[fmap[3]])))) {
	    phgDumpElement(g_, e);
	    phgDumpElement(g_, e1);
	    phgInfo(-1, "face indices of parent and child 1 mismatch\n");
	    return conformity = FALSE;
	}
    }

    return TRUE;
}

# define CheckL2Gmap(map, n, ng)

void
phgCheckConformity(GRID *g)
{
    ParallelFunction(phgCheckConformity, g != NULL && g->nprocs > 1);

    if (g == NULL) return;

    g_ = g;
    conformity = TRUE;
    phgTraverseElements(g, check_conformity_callback);

    if (conformity)
	phgTraverseAllElements(g, check_conformity_nonleaf_callback);

    if (!conformity) {
	phgError(1, "local non conformity.\n");
    }

    if (conformity) {
	if (IS_RANK0) phgInfo(1, "the grid is conforming.\n");
    }
    else {
#if 0
	phgDumpGrid(g);
#endif
	phgError(1, "%s:%d: abort.", __FILE__, __LINE__);
    }

    return;
}

#if 0
static int
elem_comp(const void *p0, const void *p1)
{
    return phgMemcmp((const SIMPLEX *)p0, (const SIMPLEX *)p1, sizeof(SIMPLEX));
}
#endif

void
phgUpdateNeighbours(GRID *g)
/* fills 'neighbours' array of all simplices. */
{
    INT i;

    FunctionEntry;

    if (g == NULL) return;

    g_ = g;

#if 0
    if (g->nprocs == 1 && g->nroot == g->nleaf) {
	/* check for duplicate tetrahedra in the initial mesh */
	INT j;
	qsort(g->roots, g->nroot, sizeof(*g->roots), elem_comp);
	j = -1;
	for (i = 0; i < g->nroot; i++) {
	    if (j < 0 || elem_comp(g->roots + j, g->roots + i) != 0)
		j++;
	    if (j != i)
		g->roots[j] = g->roots[i];
	}
	if (j < g->nroot)
	    phgWarning("duplicate tetrahedra found and removed.\n");
    }
#endif

    vertices = phgCalloc(g->nvert, sizeof(*vertices));
#if 0
    for (i = 0; i < g->nvert; i++) {
	 vertices[i].nelem = vertices[i].nmax = 0;
	 vertices[i].elements = NULL;
    }
#endif

    /* For each vertex, a list of elements to which it belongs is built */
    phgTraverseElements(g, find_neighbours_callback0);

    /* then search neighbours in the small lists of elements */
    conformity = TRUE;
    phgTraverseElements(g, find_neighbours_callback1);

    /* free memory blocks */
    for (i = 0; i < g->nvert; i++)
	if (vertices[i].elements != NULL)
	    phgFree(vertices[i].elements);
    phgFree(vertices);

    PrintTime(1);
    Return;
}

/* Note: the code for deleting degenerated elements is temperarily disabled
 * by setting REMOVE_DEGENERATED_ELEMENTS to 0 because the code crashes
 * on some bad input meshes (e.g., crask1.mesh, crask.mesh, etc.)
 *
 * It is simpler to move the code to before updating neighbour links,
 * but the present code allows transferring boundary types of deleted
 * elements. */
#define REMOVE_DEGENERATED_ELEMENTS 0

void
phgRemoveDegeneratedElements(GRID *g)
/* checks and removes degenerated elements and unused vertices.
 *
 * Note: this function should be called after updating the neighbour links,
 *	 necessary for transfering boundary types of deleted elements */
{
    INT *vmap;
    INT i, j;
    SIMPLEX *e;
#if REMOVE_DEGENERATED_ELEMENTS
    INT k, u, v, w;
    SIMPLEX *e0, *e1;
    int a, b, c;
    BOOLEAN flag;
#endif	/* REMOVE_DEGENERATED_ELEMENTS */

    FunctionEntry;

    assert(g->serial_no == 0);

    if (g->nvert <= 0)
	Return;

    vmap = phgAlloc(g->nvert * sizeof(*vmap));

    /* check and remove tetrahedra with zero length edges.
     * TODO: check for other types of degenerated tetrahedra */

#if REMOVE_DEGENERATED_ELEMENTS
    for (j = 0; j < g->nvert; j++)
	vmap[j] = j;
    flag = FALSE;
    k = 0;
    for (j = 0; j < g->nroot; j++) {
	e = g->roots + j;
	for (i = 0; i < NEdge; i++) {
	    u = e->verts[a = GetEdgeVertex(i, 0)];
	    v = e->verts[b = GetEdgeVertex(i, 1)];
	    if (g->verts[u][0] == g->verts[v][0] &&
		g->verts[u][1] == g->verts[v][1] &&
		g->verts[u][2] == g->verts[v][2])
		break;
	}
	if (i >= NEdge) {
	    /* no zero length edge in this element */
	    e->index = k++;
	    continue;
	}
	e->index = -1;	/* mark e as deleted */
	/* mark the vertex with larger value as deleted */
	w = (u > v) ? v : u;
	while (vmap[w] != w)
	    w = vmap[w];
	vmap[u] = vmap[v] = w;
	if (flag)
	    continue;
	/* update neighbour links and boundary types of the neighbours */
	e0 = e->neighbours[a];
	e1 = e->neighbours[b];
	if (e0 != NULL) {
	    for (c = 0; c < NFace; c++)
		if (e0->neighbours[c] == e)
		    break;
	    if (c >= NVert) {
		flag = TRUE;
		continue;
	    }
	    e0->neighbours[c] = e1;
	    e0->bound_type[c] = e->bound_type[b];
	}
	if (e1 != NULL) {
	    for (c = 0; c < NFace; c++)
		if (e1->neighbours[c] == e)
		    break;
	    if (c >= NVert) {
		flag = TRUE;
		continue;
	    }
	    e1->neighbours[c] = e0;
	    e1->bound_type[c] = e->bound_type[a];
	}
    }

    /* update vmap[] */
    if (k < g->nroot) {
	for (j = 0; j < g->nvert; j++) {
	    v = j;
	    while (v != vmap[v])
		v = vmap[v];
	    vmap[j] = v;
	}
    }

    /* remove tetrahedra with duplicate vertices */
    k = 0;
    for (j = 0; j < g->nroot; j++) {
	INT verts[NVert];
	e = g->roots + j;
	if (e->index == -1)
	    continue;
	for (i = 0; i < NVert; i++)
	    verts[i] = e->verts[i] = vmap[e->verts[i]];
	qsort(verts, NVert, sizeof(*verts), phgCompINT);
	if (verts[0] < 0 || verts[NVert - 1] >= g->nvert) {
	    flag = TRUE;
	    e->index = -1;
	    continue;
	}
	for (i = 1; i < NVert; i++) {
	    if (verts[i] == verts[i - 1])
		break;
	}
	if (i < NVert) {
	    flag = TRUE;
	    e->index = -1;
	    continue;
	}
	e->index = k++;
    }

    if (k < g->nroot) {
	/* some elements are to be deleted, update neighbours and vertices */
	i = g->nroot - k;
	phgInfo(2, "deleting %d degenerated element%s\n", i, i > 1 ? "s" : "");
	for (j = 0; j < g->nroot; j++) {
	    e = g->roots + j;
	    if (e->index == -1)
		continue;
	    for (i = 0; i < NVert; i++) {
		if ((e0 = e->neighbours[i]) != NULL) {
		    if (!flag) {
			assert(e0->index >= 0);
			e->neighbours[i] = g->roots + (e0->index);
		    }
		    else {
			e->neighbours[i] = NULL;
		    }
		}
	    }
	}
	for (j = k = 0; j < g->nroot; j++) {
	    if (g->roots[j].index == -1)
		continue;
	    if (j != k)
		g->roots[k] = g->roots[j];
	    k++;
	}
	g->ntree = g->nleaf = g->nelem = g->nroot = g->nelem_global = k;
	if (flag) {
	    phgWarning("bdry type of deleted elements may not be "
			"correctly transferred.\n");
	    phgUpdateNeighbours(g);
	}
    }
#endif	/* REMOVE_DEGENERATED_ELEMENTS */

    /* remove unused vertices, they may either come from the imported file
     * or from the deleted elements */

    memset(vmap, 0, g->nvert * sizeof(*vmap));
    for (j = 0; j < g->nroot; j++) {
	e = g->roots + j;
	for (i = 0; i < NVert; i++)
	    vmap[e->verts[i]] = 1;
    }

    /* compute new index in vmap and update g->verts */
    j = 0;
    for (i = 0; i < g->nvert; i++) {
	if (vmap[i] == 0) {
	    /* vertex unused */
	    vmap[i] = -1;
	    continue;
	}
	vmap[i] = j;
	if (j != i)
	    memcpy(g->verts + j, g->verts + i, sizeof(COORD));
	j++;
    }

    if (j == g->nvert) {
	/* no unused vertices */
	phgFree(vmap);
	Return;
    }

    i = g->nvert - j;
    phgInfo(2, "deleting %d unused vert%s\n", i, i > 1 ? "ices" : "ex");
    g->nvert = g->nvert_global = j;
    if (j == 0)
	g->verts = NULL;
    else {
	g->verts = phgRealloc(g->verts, j * sizeof(*g->verts));
	for (j = 0; j < g->nroot; j++) {
	    e = g->roots + j;
	    for (i = 0; i < NVert; i++)
		e->verts[i] = vmap[e->verts[i]];
	}
    }

    phgFree(vmap);
    Return;
}

static SIMPLEX *
find_edge_neighbour(SIMPLEX *e, INT u0, INT u1, INT *u)
/* Returns the neighbour not containing the face {u0,u1,u}, but containing
 * the edge u0-u1, where u0, u1, *u are global vetex indices. On return
 * *u is set to the local index of the third vertex of the common face. */
{
    int i, j;

    /* one of the vertices must be equal to "*u" */
    if (e->verts[i = 0] != *u && e->verts[++i] != *u &&
	e->verts[++i] != *u && e->verts[++i] != *u) {
	phgInfo(-1, "u0=%d, u1=%d, *u=%d\n", u0, u1, *u);
	phgDumpElement(NULL, e);
	phgError(1, "%s:%d, unexpected error!\n", __FILE__, __LINE__);
    }

    for (j = 0; j == i || e->verts[j] == u0 || e->verts[j] == u1; j++);
    *u = e->verts[j];

    /* it is assumed that e->neighbours[i] == NULL for all non-interior and
     * REMOTE neighbours */
    return IsLeaf(e) ? GetNeighbour(e, i) : e->neighbours[i];
}

INT
phgGetPatch(GRID *g, SIMPLEX *e, BYTE v0, BYTE v1, SIMPLEX ***p)
/* constructs the list of elements constituting the patch around
 * the edge v0-v1 (local indices) of e. returns number of elements
 * in the patch and list of elements through 'p' */
/* TODO: patch composed of several non contiguous parts */
{
    INT u0 = e->verts[v0], u1 = e->verts[v1], u;
    /* gcc is unhappy if v2, v3 are not initialized :( */
    BYTE v2 = 0, v3 = 0;
    INT count = 0, size;
    SIMPLEX *e1;

    *p = phgAlloc((size = 6) * sizeof(**p));
    (*p)[count++] = e;

    /* find out the two opposite vertices, v2 and v3, of the edge */
    switch (v0 * 4 + v1) {
	case 0 * 4 + 1: v2 = 2; v3 = 3; break;
	case 0 * 4 + 2: v2 = 1; v3 = 3; break;
	case 0 * 4 + 3: v2 = 1; v3 = 2; break;
	case 1 * 4 + 2: v2 = 0; v3 = 3; break;
	case 1 * 4 + 3: v2 = 0; v3 = 2; break;
	case 2 * 4 + 3: v2 = 0; v3 = 1; break;
    }
    
    /* Note: refine_patch depends on the direction, don't change! */

    /* search in first direction */
    e1 = IsLeaf(e) ? GetNeighbour(e, v2) : e->neighbours[v2];
    u = e->verts[v3];
    while (e1 != e && e1 != NULL) {
	if (count >= size) {
	    if (count > 100)
		phgError(1, "%s:%d, unexpected error.\n", __FILE__, __LINE__);
	    *p = phgRealloc(*p, (size += 6) * sizeof(**p));
	}
	(*p)[count++] = e1;
	e1 = find_edge_neighbour(e1, u0, u1, &u);
    }
    if (e1 == e) {
	phgDebug((4, "circular patch, size = %d\n", count));
	return count;
    }
    /* search in second direction */
#if 0
{
    phgInt count0 = count;	/* save position */
#endif
    e1 = IsLeaf(e) ? GetNeighbour(e, v3) : e->neighbours[v3];
    u = e->verts[v2];
    /* Note: e1 may come back to e when dealing with non-leaf elements
     * on a distributed mesh */
    while (e1 != e && e1 != NULL) {
	if (count >= size) {
	    if (count > 100)
		phgError(1, "%s:%d, unexpected error.\n", __FILE__, __LINE__);
	    *p = phgRealloc(*p, (size += 6) * sizeof(**p));
	}
	(*p)[count++] = e1;
	e1 = find_edge_neighbour(e1, u0, u1, &u);
#if DEBUG
	if (e1 == e) {
	    int i;
	    phgInfo(-1, "unexpected case: e1 == e,\n");
	    for (i = 0; i < count; i++)
		phgDumpElement(g, (*p)[i]);
	    phgError(1, "abort.\n");
	}
#endif
    }

#if 0
    /* We have the list of elements around the patch in the following order:
     *		*p[0] -> *p[1] -> ... -> *p[count0-1]
     *		*p[0] <- *p[count0] <- ... <- *p[count-1]
     *
     * We first commute (*p[0], *p[count0-1]), (*p[1], *p[count0-2]), etc.,
     * the we commute (*p[0], *p[count-1]), (*p[1], *p[count-2]), etc.,
     * such that the elements in the list are chained in the right direction */
    for (u0 = 0, u1 = count0 - 1; u0 < u1; u0++, u1--) {
	e = (*p)[u0];
	(*p)[u0] = (*p)[u1];
	(*p)[u1] = e;
    }
    for (u0 = 0, u1 = count - 1; u0 < u1; u0++, u1--) {
	e = (*p)[u0];
	(*p)[u0] = (*p)[u1];
	(*p)[u1] = e;
    }
}
#endif

    phgDebug((4, "open patch, size = %d\n", count));
    return count;    
}

static BOOLEAN
update_edges_callback CB_ARGS(e)
{
    int i, j, k, count;
    SIMPLEX **p, *e1;
    BOOLEAN flag;	/* indicate whether a shared edge */

    /* loop on all edges of the element */
    for (i = 0; i < NEdge; i++) {
	int v0 = GetEdgeVertex(i, 0), v1 = GetEdgeVertex(i, 1);
	INT g0, g1;	/* global indices of the two vertices */
	if (e->edges[i] != -1)
	    continue;
	/* new edge, assign it to all elements sharing it */
	g0 = e->verts[v0];
	g1 = e->verts[v1];
	/* loop on elements in the patch */
	count = phgGetPatch(g_, e, v0, v1, &p);
	flag = FALSE;
	for (j = 0; j < count; j++) {
	    /* loop on edges of p[j] to find the one matching current edge */
	    e1 = p[j];
	    for (k = 0; k < NEdge; k++) {
		INT h0, h1;
		if (e1->edges[k] != -1) continue;
		h0 = e1->verts[GetEdgeVertex(k, 0)];
		h1 = e1->verts[GetEdgeVertex(k, 1)];
		if ((h0 == g0 && h1 == g1) || (h0 == g1 && h1 == g0)) break;
	    }
	    if (k >= NEdge) {
		phgDumpElement(g_, e1);
		phgInfo(-1, "common edge: %d %d\n", GlobalVertex(g_, g0),
			GlobalVertex(g_, g1));
		phgError(1, "%s:%d: unexpected error.\n", __FILE__, __LINE__);
	    }
	    e1->edges[k] = g_->nedge;
	}
	phgFree(p);
	g_->nedge++;
    }
    return TRUE;
}

void
phgUpdateEdges(GRID *g)
{
    FunctionEntry;

    if (g == NULL) return;

    /* traverse all edges */
    phgDebug((3, "collecting edges\n"));
    g_ = g;
    phgTraverseElements(g, update_edges_callback);

    g->nedge_global = g->nedge;

    PrintTime(1);
    phgInfo(2, "number of edges = %d\n", g->nedge);

    Return;
}

#if Dim == 3
static INT faces_local, faces_shared, gindex;

static BOOLEAN
update_faces_callback0 CB_ARGS(e)
{
    int i;

    /* loop on all faces of the element */
    for (i = 0; i < NFace; i++) {
	if (e->faces[i] != -1) continue;
	if (HasLocalNeighbour(e, i)) {
	    SIMPLEX *e1 = e->neighbours[i];
	    if (e->verts[i] < e1->verts[phgOppositeVertex(e, i, e1)])
		faces_local++;
	}
	else if (!HasRemoteNeighbour(e, i)) {
	    faces_local++;
	}
	else {
	    faces_shared++;
	}
    }
    return TRUE;
}

static BOOLEAN
update_faces_callback1 CB_ARGS(e)
{
    int i, j;
    SIMPLEX *e1;

    /* loop on all faces of the element */
    for (i = 0; i < NFace; i++) {
	if (e->faces[i] != -1 || (e->bound_type[i] & REMOTE))
	    continue;
	if (HasLocalNeighbour(e, i)) {
	    j = phgOppositeVertex(e, i, e1 = e->neighbours[i]);
	    if (e->verts[i] < e1->verts[j]) {
		e->faces[i] = e1->faces[j] = g_->nface++;
		gindex++;
	    }
	}
	else {
	    e->faces[i] = g_->nface++;
	    gindex++;
	}
    }
    return TRUE;
}

void
phgUpdateFaces(GRID *g)
{
    FunctionEntry;

    if (g == NULL) Return;
    g_ = g;

    /* count number of unassigned faces (those having index -1) */
    faces_local = faces_shared = 0;
    phgTraverseElements(g, update_faces_callback0);
    phgDebug((3, "faces: local=%d, shared=%d\n", faces_local, faces_shared));
    gindex = g->nface_global;

    /* assign local indices */
    phgTraverseElements(g, update_faces_callback1);

    if (g->nprocs <= 1)
	g->nface_global = g->nface;

    PrintTime(1);
    Return;
}
#endif	/* Dim == 3 */

void
phgUpdateElementIndices(GRID *g, INT nelem_old, INT nelem_global_old)
{

    FunctionEntry;

    (void)g;
    (void)nelem_old;
    (void)nelem_global_old;

    Return;
}

void
phgDumpElement(GRID *g, SIMPLEX *e)
{
    int i;
    char s[128];
#ifdef VL_
# undef VL_
#endif
#define VL_(i)	(e->verts[i])
#ifdef EL_
# undef EL_
#endif
#define EL_(i)	(e->edges[i])
#ifdef FL_
# undef FL_
#endif
#define FL_(i)	(e->faces[i])

    if (e == NULL) {
	phgWarning("phgDumpElement: e == NULL\n", e);
	return;
    }

#if Dim == 2
    phgInfo(-1, "E(%d): %p, %s, mark=%d, bdry=%x:%x:%x, " "gen=%d\n",
		g->elem_flag ? GlobalElement(g, e->index) : -1,
		e, GTypeName(e->type), e->mark,
		e->bound_type[0], e->bound_type[1], e->bound_type[2],
		e->generation);
    phgInfo(-1, "  children=%p %p, parent=%p\n",
		e->children[0], e->children[1]);
    phgInfo(-1, "  verts=%d %d %d\n", VL_(0), VL_(1), VL_(2));
    phgInfo(-1, "  edges=%d %d %d\n", e->edges[0], e->edges[1], e->edges[2]);
#else	/* Dim */
    phgInfo(-1, "E(%d): %p, %s, mark=%d, bdry=%x:%x:%x:%x, " "gen=%d\n",
		g->elem_flag ? GlobalElement(g, e->index) : -1,
		e, GTypeName(e->type), e->mark,
		e->bound_type[0], e->bound_type[1], e->bound_type[2],
		e->bound_type[3], e->generation);
    phgInfo(-1, "  children=%p %p, parent=%p\n",
		e->children[0], e->children[1]);
    phgInfo(-1, "  verts=%d %d %d %d\n", VL_(0), VL_(1), VL_(2), VL_(3));
    phgInfo(-1, "  edges=%d %d %d %d %d %d\n", e->edges[0], e->edges[1],
		e->edges[2], e->edges[3], e->edges[4], e->edges[5]);
    phgInfo(-1, "  faces=%d %d %d %d\n", e->faces[0], e->faces[1],
		e->faces[2], e->faces[3]);
#endif	/* Dim */
#undef V_
    s[0] = '\0';
    for (i = 0; i < NFace; i++) {
	if (HasLocalNeighbour(e, i))
	    sprintf(s+strlen(s), "%p ", e->neighbours[i]);
	else if ((e->bound_type[i] & DIRICHLET))
	    sprintf(s+strlen(s), "D ");
	else if ((e->bound_type[i] & NEUMANN))
	    sprintf(s+strlen(s), "N ");
	else if ((e->bound_type[i] & UNDEFINED))
	    sprintf(s+strlen(s), "U ");
	else
	    sprintf(s+strlen(s), "undef ");
    }
    phgInfo(-1, "  neighbours=%s\n", s);

    if (phgVerbosity < 2) return;

}

void
phgDumpPatch(GRID *g, SIMPLEX *e, int v0, int v1)
/* dumps out a patch */
{
    int j, n;
    SIMPLEX **p;

    phgInfo(-1, "========= the patch:\n");
    n = phgGetPatch(g, e, v0, v1, &p);
    for (j = 0; j < n; j++) phgDumpElement(g, p[j]);
    phgFree(p);
    phgInfo(-1, "=========\n");
}

static BOOLEAN
dump_callback CB_ARGS(e)
{
    phgDumpElement(g_, e);
    return TRUE;
}

static void
dump_head(GRID *g)
{
    phgInfo(-1, "vvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvv\n");
    phgInfo(-1, "nroot = %d\n", g->nroot);
    phgInfo(-1, "nleaf = %d\n", g->nleaf);
    phgInfo(-1, "ntree = %d\n", g->ntree);
    phgInfo(-1, "nvert = %d\n", g->nvert);
    phgInfo(-1, "nedge = %d\n", g->nedge);
#if Dim == 3
    phgInfo(-1, "nface = %d\n", g->nface);
#endif
    phgInfo(-1, "nelem = %d\n", g->nelem);
    /*if (g->nprocs > 1)*/ {
	phgInfo(-1, "nvert_global = %d\n", g->nvert_global);
	phgInfo(-1, "nedge_global = %d\n", g->nedge_global);
#if Dim == 3
	phgInfo(-1, "nface_global = %d\n", g->nface_global);
#endif
	phgInfo(-1, "nelem_global = %d\n", g->nelem_global);
    }
}

static void
dump_tail(GRID *g)
{
    Unused(g);
    phgInfo(-1, "^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^\n");
}

void
phgDumpGridInfo(GRID *g)
{
    ParallelFunction(phgDumpGridInfo, g != NULL && g->nprocs > 1);
    if (g == NULL) return;
    dump_head(g_ = g);
    dump_tail(g);
}

void
phgDumpGrid(GRID *g)
{
    ParallelFunction(phgDumpGrid, g != NULL && g->nprocs > 1);
    if (g == NULL) return;
    dump_head(g_ = g);
    phgTraverseElements(g, dump_callback);
    dump_tail(g);
}

void
phgDumpTree(GRID *g)
{
    ParallelFunction(phgDumpTree, g != NULL && g->nprocs > 1);
    if (g == NULL) return;
    dump_head(g_ = g);
    phgTraverseAllElements(g, dump_callback);
    dump_tail(g);
}

void
phgDumpBranch(GRID *g, SIMPLEX *e)
{
    phgDumpElement(g, e);
    if (e->children[0] != NULL)
	phgDumpBranch(g, e->children[0]);
    if (e->children[1] != NULL)
	phgDumpBranch(g, e->children[1]);
}

#if defined(GZIP_PROG) || defined(BZIP2_PROG)
#define MAX_FILES 16
static struct {
    FILE *fp;
    BOOLEAN is_pipe;
} open_files[MAX_FILES];
#endif

void
phgCloseInputFile_(FILE *fp)
{
#if defined(GZIP_PROG) || defined(BZIP2_PROG)
    int i;
    BOOLEAN is_pipe = FALSE;
    for (i = 0; i < MAX_FILES; i++) {
	if (open_files[i].fp == fp) {
	    is_pipe = open_files[i].is_pipe;
	    open_files[i].fp = NULL;
	    break;
	}
    }
    (is_pipe ? pclose : fclose)(fp);
#else
    if (fp != NULL) fclose(fp);
#endif
}

FILE *
phgOpenInputFile_(const char *filename)
{
    FILE *fp;
#if defined(GZIP_PROG) || defined(BZIP2_PROG)
    char cmd[PATH_MAX + 16];
    int len = strlen(filename);
    BOOLEAN is_pipe = FALSE;
    static BOOLEAN initialized = FALSE;

    if (!initialized) {
	int i;
	initialized = TRUE;
	for (i = 0; i < MAX_FILES; i++) open_files[i].fp = NULL;
    }
#endif

#ifdef GZIP_PROG
    if (len > 3 && !phgMemcmp(filename + len - 3, ".gz", 3)) {
	if ((fp = fopen(filename, "r")) != NULL) {
	    fclose(fp);
	    phgInfo(1, "gzipped file.\n");
	    sprintf(cmd, GZIP_PROG " -dc %s 2>/dev/null", filename);
	    fp = popen(cmd, "r");
	    is_pipe = TRUE;
	}
    } else
#endif
#ifdef BZIP2_PROG
    if (len > 4 && !phgMemcmp(filename + len - 4, ".bz2", 4)) {
	if ((fp = fopen(filename, "r")) != NULL) {
	    fclose(fp);
	    phgInfo(1, "bzip2 file.\n");
	    sprintf(cmd, BZIP2_PROG " -dc %s 2>/dev/null", filename);
	    fp = popen(cmd, "r");
	    is_pipe = TRUE;
	}
    } else
#endif
    {
	fp = fopen(filename, "r");
    }
    if (fp == NULL)
	phgWarning("can't open file \"%s\"!\n", filename);
#if defined(GZIP_PROG) || defined(BZIP2_PROG)
    else {
	int i;
	for (i = 0; i < MAX_FILES; i++)
	    if (open_files[i].fp == NULL) break;
	if (i >= MAX_FILES) {
	    phgWarning("too many open files.\n");
	    (is_pipe ? pclose : fclose)(fp);
	    fp = NULL;
	}
	else {
	    open_files[i].fp = fp;
	    open_files[i].is_pipe = is_pipe;
	}
    }
#endif

    return fp;
}

BDRY_MAP_FUNC phgBdryMapFunc_ = NULL;

void
phgImportSetBdryMapFunc(BDRY_MAP_FUNC func)
{
    phgBdryMapFunc_ = func;
}

BOOLEAN
phgImport(GRID *g, const char *filename, BOOLEAN distr)
{
    FILE *fp;
    char line[128], magic[128];
    BOOLEAN ret = FALSE;

    if (g == NULL)
	phgError(1, "invalid grid pointer.\n");

    if (g->nleaf_global > 0) {
	phgWarning("g already contains a valid grid, new grid not imported.\n");
	return ret;
    }

    if (phgRank > 0)
	goto end;

    if ((fp = phgOpenInputFile_(filename)) == NULL) {
	phgError(1, "cannot read file \"%s\".\n", filename);
	goto end;
    }

    if (fgets(line, sizeof(line), fp) == NULL ||
	sscanf(line, "%s", magic) != 1) {
	phgCloseInputFile_(fp);
	goto end;
    }
    phgCloseInputFile_(fp);

    if (!strncmp(magic, "DIM:", 4) || !strncmp(magic, "DIM_OF_WORLD:", 13))
	ret = phgImportALBERT(g, filename, FALSE);
    else if (!strcmp(magic, "MeshVersionFormatted"))
	ret = phgImportMedit(g, filename, FALSE);
    else {
	phgWarning("can't determine type of input file, mesh not imported.\n");
    }

    if (ret && phgRank == 0)
	phgInfo(1, "%d elements loaded.\n", g->nleaf);

end:

    if (g->geom_flag)
	phgGeomInit(g);

    if (ret)
	g->serial_no++;

    return ret;
}

#if ALLOW_CURVED_BOUNDARY

void phgInitBoundaryFunctions(GRID *g)
{
    EXPR **funcs;
    BYTE *flags;
    SIMPLEX *e;
    int i, j;

    if ((funcs = g->bdry_funcs) == NULL)
	return;

    flags = phgCalloc(g->nvert, sizeof(*flags));

    /* mark boundary vertices (bit 0) */
    for (i = 0; i < g->nroot; i++) {
	e = g->roots + i;
	for (j = 0; j < NFace; j++) {
	    if (e->bound_type[j] & BDRY_MASK) {
		int k;
		for (k = 0; k < NFace; k++)
		    if (k != j)
			flags[e->verts[k]] = 1;
	    }
	}
    }

    while (*funcs != NULL) {
	for (j = 0; j < g->nvert; j++) {
	    COORD *c;
	    if (!flags[j])
		continue;
	    c = g->verts + j;
	    if (fabs(phgEvaluate3DFunction(*funcs, (*c)[0], (*c)[1], (*c)[2],
			0.0)) < 1e-4)
		flags[j] |= 2;
	    else
		flags[j] &= 1;
	}
	for (i = 0; i < g->nroot; i++) {
	    e = g->roots + i;
	    for (j = 0; j < NEdge; j++) {
		if (e->bound_func[j] != -1)
		    continue;
		if ((flags[e->verts[GetEdgeVertex(j, 0)]] & 2) != 0 &&
		    (flags[e->verts[GetEdgeVertex(j, 1)]] & 2) != 0)
		    e->bound_func[j] = (funcs - g->bdry_funcs) / 4;
	    }
	}
	funcs += 4;
    }

    phgFree(flags);
}

FLOAT
phgGetBoundaryError(GRID *g, SIMPLEX *e)
/* returns the maximum square of the distances between the middle points of
 * the edges and their projections onto curved boundary */
{
    EXPR **f;
    COORD *v0, *v1;
    FLOAT d, dmax, x, y, z, dx, dy, dz;
    int edge;

    dmax = 0.0;
    for (edge = 0; edge < NEdge; edge++) {
	if (e->bound_func[edge] == -1)
	    continue;
    
	f = g->bdry_funcs + e->bound_func[edge] * 4;
	v0 = g->verts + e->verts[GetEdgeVertex(edge, 0)];
	v1 = g->verts + e->verts[GetEdgeVertex(edge, 1)];
	x = ((*v0)[0] + (*v1)[0]) * .5;
	y = ((*v0)[1] + (*v1)[1]) * .5;
	z = ((*v0)[2] + (*v1)[2]) * .5;
	dx = phgEvaluate3DFunction(f[1], x, y, z, 0.0) - x;
	dy = phgEvaluate3DFunction(f[2], x, y, z, 0.0) - y;
	dz = phgEvaluate3DFunction(f[3], x, y, z, 0.0) - z;
	/* FIXME: sum or max? */
	d = dx * dx + dy * dy + dz * dz;
	if (dmax < d)
	    dmax = d;
    }

    return dmax;
}

#else	/* ALLOW_CURVED_BOUNDARY */

FLOAT
phgGetBoundaryError(GRID *g, SIMPLEX *e)
{
    Unused(g);
    Unused(e);
    return 0.0;
}

#endif	/* ALLOW_CURVED_BOUNDARY */

